# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing for conversions."""

from typing import TYPE_CHECKING

from pint import Quantity, UndefinedUnitError

from ansys.api.geometry.v0.models_pb2 import (
    Arc as GRPCArc,
    Circle as GRPCCircle,
    CurveGeometry as GRPCCurve,
    Direction as GRPCDirection,
    Ellipse as GRPCEllipse,
    Frame as GRPCFrame,
    Geometries as GRPCGeometries,
    Line as GRPCLine,
    Material as GRPCMaterial,
    MaterialProperty as GRPCMaterialProperty,
    Matrix as GRPCMatrix,
    Plane as GRPCPlane,
    Point as GRPCPoint,
    Polygon as GRPCPolygon,
    Surface as GRPCSurface,
    TrimmedCurve as GRPCTrimmedCurve,
)
from ansys.geometry.core.materials.material import (
    Material,
    MaterialProperty,
    MaterialPropertyType,
)
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point2D, Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, UNITS
from ansys.geometry.core.shapes.curves.circle import Circle
from ansys.geometry.core.shapes.curves.curve import Curve
from ansys.geometry.core.shapes.curves.ellipse import Ellipse
from ansys.geometry.core.shapes.curves.line import Line
from ansys.geometry.core.shapes.surfaces.cone import Cone
from ansys.geometry.core.shapes.surfaces.cylinder import Cylinder
from ansys.geometry.core.shapes.surfaces.plane import PlaneSurface
from ansys.geometry.core.shapes.surfaces.sphere import Sphere
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.torus import Torus
from ansys.geometry.core.sketch.arc import Arc
from ansys.geometry.core.sketch.circle import SketchCircle
from ansys.geometry.core.sketch.edge import SketchEdge
from ansys.geometry.core.sketch.ellipse import SketchEllipse
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.polygon import Polygon
from ansys.geometry.core.sketch.segment import SketchSegment

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.face import SurfaceType
    from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve


def unit_vector_to_grpc_direction(unit_vector: UnitVector3D) -> GRPCDirection:
    """Convert a ``UnitVector3D`` class to a unit vector gRPC message.

    Parameters
    ----------
    unit_vector : UnitVector3D
        Source vector data.

    Returns
    -------
    GRPCDirection
        Geometry service gRPC direction message.
    """
    return GRPCDirection(x=unit_vector.x, y=unit_vector.y, z=unit_vector.z)


def frame_to_grpc_frame(frame: Frame) -> GRPCFrame:
    """Convert a ``Frame`` class to a frame gRPC message.

    Parameters
    ----------
    frame : Frame
        Source frame data.

    Returns
    -------
    GRPCFrame
        Geometry service gRPC frame message. The unit for the frame origin is meters.
    """
    return GRPCFrame(
        origin=point3d_to_grpc_point(frame.origin),
        dir_x=unit_vector_to_grpc_direction(frame.direction_x),
        dir_y=unit_vector_to_grpc_direction(frame.direction_y),
    )


def plane_to_grpc_plane(plane: Plane) -> GRPCPlane:
    """Convert a ``Plane`` class to a plane gRPC message.

    Parameters
    ----------
    plane : Plane
        Source plane data.

    Returns
    -------
    GRPCPlane
        Geometry service gRPC plane message. The unit is meters.
    """
    return GRPCPlane(
        frame=GRPCFrame(
            origin=point3d_to_grpc_point(plane.origin),
            dir_x=unit_vector_to_grpc_direction(plane.direction_x),
            dir_y=unit_vector_to_grpc_direction(plane.direction_y),
        )
    )


def sketch_shapes_to_grpc_geometries(
    plane: Plane,
    edges: list[SketchEdge],
    faces: list[SketchFace],
    only_one_curve: bool = False,
) -> GRPCGeometries:
    """Convert lists of ``SketchEdge`` and ``SketchFace`` to a gRPC message.

    Parameters
    ----------
    plane : Plane
        Plane for positioning the 2D sketches.
    edges : list[SketchEdge]
        Source edge data.
    faces : list[SketchFace]
        Source face data.
    only_one_curve : bool, default: False
        Whether to project one curve of the whole set of geometries to
        enhance performance.

    Returns
    -------
    GRPCGeometries
        Geometry service gRPC geometries message. The unit is meters.
    """
    geometries = GRPCGeometries()

    converted_sketch_edges = sketch_edges_to_grpc_geometries(edges, plane)
    geometries.lines.extend(converted_sketch_edges[0])
    geometries.arcs.extend(converted_sketch_edges[1])

    for face in faces:
        if isinstance(face, SketchCircle):
            geometries.circles.append(sketch_circle_to_grpc_circle(face, plane))
        if isinstance(face, SketchEllipse):
            geometries.ellipses.append(sketch_ellipse_to_grpc_ellipse(face, plane))
        if isinstance(face, Polygon):
            geometries.polygons.append(sketch_polygon_to_grpc_polygon(face, plane))
        else:
            converted_face_edges = sketch_edges_to_grpc_geometries(face.edges, plane)
            geometries.lines.extend(converted_face_edges[0])
            geometries.arcs.extend(converted_face_edges[1])

    if only_one_curve:
        one_curve_geometry = GRPCGeometries()
        if len(geometries.lines) > 0:
            one_curve_geometry.lines.append(geometries.lines[0])
        elif len(geometries.arcs) > 0:
            one_curve_geometry.arcs.append(geometries.arcs[0])
        elif len(geometries.circles) > 0:
            one_curve_geometry.circles.append(geometries.circles[0])
        elif len(geometries.ellipses) > 0:
            one_curve_geometry.ellipses.append(geometries.ellipses[0])
        elif len(geometries.polygons) > 0:
            one_curve_geometry.polygons.append(geometries.polygons[0])
        return one_curve_geometry

    else:
        return geometries


def sketch_edges_to_grpc_geometries(
    edges: list[SketchEdge],
    plane: Plane,
) -> tuple[list[GRPCLine], list[GRPCArc]]:
    """Convert a list of ``SketchEdge`` to a gRPC message.

    Parameters
    ----------
    edges : list[SketchEdge]
        Source edge data.
    plane : Plane
        Plane for positioning the 2D sketches.

    Returns
    -------
    tuple[list[GRPCLine], list[GRPCArc]]
        Geometry service gRPC line and arc messages. The unit is meters.
    """
    arcs = []
    segments = []
    for edge in edges:
        if isinstance(edge, SketchSegment):
            segments.append(sketch_segment_to_grpc_line(edge, plane))
        elif isinstance(edge, Arc):
            arcs.append(sketch_arc_to_grpc_arc(edge, plane))

    return (segments, arcs)


def sketch_arc_to_grpc_arc(arc: Arc, plane: Plane) -> GRPCArc:
    """Convert an ``Arc`` class to an arc gRPC message.

    Parameters
    ----------
    arc : Arc
        Source arc data.
    plane : Plane
        Plane for positioning the arc within.

    Returns
    -------
    GRPCArc
        Geometry service gRPC arc message. The unit is meters.
    """
    axis = (
        unit_vector_to_grpc_direction(plane.direction_z)
        if not arc.is_clockwise
        else unit_vector_to_grpc_direction(-plane.direction_z)
    )

    return GRPCArc(
        center=point2d_to_grpc_point(plane, arc.center),
        start=point2d_to_grpc_point(plane, arc.start),
        end=point2d_to_grpc_point(plane, arc.end),
        axis=axis,
    )


def sketch_ellipse_to_grpc_ellipse(ellipse: SketchEllipse, plane: Plane) -> GRPCEllipse:
    """Convert a ``SketchEllipse`` class to an ellipse gRPC message.

    Parameters
    ----------
    ellipse : SketchEllipse
        Source ellipse data.

    Returns
    -------
    GRPCEllipse
        Geometry service gRPC ellipse message. The unit is meters.
    """
    return GRPCEllipse(
        center=point2d_to_grpc_point(plane, ellipse.center),
        majorradius=ellipse.major_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        minorradius=ellipse.minor_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        angle=ellipse.angle.m_as(DEFAULT_UNITS.SERVER_ANGLE),
    )


def sketch_circle_to_grpc_circle(circle: SketchCircle, plane: Plane) -> GRPCCircle:
    """Convert a ``SketchCircle`` class to a circle gRPC message.

    Parameters
    ----------
    circle : SketchCircle
        Source circle data.
    plane : Plane
        Plane for positioning the circle.

    Returns
    -------
    GRPCCircle
        Geometry service gRPC circle message. The unit is meters.
    """
    return GRPCCircle(
        center=point2d_to_grpc_point(plane, circle.center),
        radius=circle.radius.m_as(DEFAULT_UNITS.SERVER_LENGTH),
    )


def point3d_to_grpc_point(point: Point3D) -> GRPCPoint:
    """Convert a ``Point3D`` class to a point gRPC message.

    Parameters
    ----------
    point : Point3D
        Source point data.

    Returns
    -------
    GRPCPoint
        Geometry service gRPC point message. The unit is meters.
    """
    return GRPCPoint(
        x=point.x.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        y=point.y.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        z=point.z.m_as(DEFAULT_UNITS.SERVER_LENGTH),
    )


def grpc_point_to_point3d(point: GRPCPoint) -> Point3D:
    """Convert a point gRPC message class to a ``Point3D`` class.

    Parameters
    ----------
    point : GRPCPoint
        Source point data.

    Returns
    -------
    Point3D
        Converted point.
    """
    return Point3D(
        [point.x, point.y, point.z],
        DEFAULT_UNITS.SERVER_LENGTH,
    )


def point2d_to_grpc_point(plane: Plane, point2d: Point2D) -> GRPCPoint:
    """Convert a ``Point2D`` class to a point gRPC message.

    Parameters
    ----------
    plane : Plane
        Plane for positioning the 2D point.
    point : Point2D
        Source point data.

    Returns
    -------
    GRPCPoint
        Geometry service gRPC point message. The unit is meters.
    """
    point3d = plane.transform_point2d_local_to_global(point2d)
    return GRPCPoint(
        x=point3d.x.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        y=point3d.y.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        z=point3d.z.m_as(DEFAULT_UNITS.SERVER_LENGTH),
    )


def sketch_polygon_to_grpc_polygon(polygon: Polygon, plane: Plane) -> GRPCPolygon:
    """Convert a ``Polygon`` class to a polygon gRPC message.

    Parameters
    ----------
    polygon : Polygon
        Source polygon data.

    Returns
    -------
    GRPCPolygon
        Geometry service gRPC polygon message. The unit is meters.
    """
    return GRPCPolygon(
        center=point2d_to_grpc_point(plane, polygon.center),
        radius=polygon.inner_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        numberofsides=polygon.n_sides,
        angle=polygon.angle.m_as(DEFAULT_UNITS.SERVER_ANGLE),
    )


def sketch_segment_to_grpc_line(segment: SketchSegment, plane: Plane) -> GRPCLine:
    """Convert a ``Segment`` class to a line gRPC message.

    Parameters
    ----------
    segment : SketchSegment
        Source segment data.

    Returns
    -------
    GRPCLine
        Geometry service gRPC line message. The unit is meters.
    """
    return GRPCLine(
        start=point2d_to_grpc_point(plane, segment.start),
        end=point2d_to_grpc_point(plane, segment.end),
    )


def grpc_matrix_to_matrix(m: GRPCMatrix) -> Matrix44:
    """Convert an ``ansys.api.geometry.Matrix`` to a ``Matrix44``."""
    import numpy as np

    return Matrix44(
        np.round(
            [
                [m.m00, m.m01, m.m02, m.m03],
                [m.m10, m.m11, m.m12, m.m13],
                [m.m20, m.m21, m.m22, m.m23],
                [m.m30, m.m31, m.m32, m.m33],
            ],
            8,
        )
    )


def grpc_frame_to_frame(frame: GRPCFrame) -> Frame:
    """Convert a frame gRPC message to a ``Frame`` class.

    Parameters
    ----------
    GRPCFrame
        Geometry service gRPC frame message. The unit for the frame origin is meters.

    Returns
    -------
    frame : Frame
        Resulting converted frame.
    """
    return Frame(
        Point3D(
            [
                frame.origin.x,
                frame.origin.y,
                frame.origin.z,
            ],
            DEFAULT_UNITS.SERVER_LENGTH,
        ),
        UnitVector3D(
            [
                frame.dir_x.x,
                frame.dir_x.y,
                frame.dir_x.z,
            ]
        ),
        UnitVector3D(
            [
                frame.dir_y.x,
                frame.dir_y.y,
                frame.dir_y.z,
            ]
        ),
    )


def grpc_surface_to_surface(surface: GRPCSurface, surface_type: "SurfaceType") -> Surface:
    """Convert a surface gRPC message to a ``Surface`` class.

    Parameters
    ----------
    surface : GRPCSurface
        Geometry service gRPC surface message.

    Returns
    -------
    Surface
        Resulting converted surface.
    """
    from ansys.geometry.core.designer.face import SurfaceType

    origin = Point3D(
        [surface.origin.x, surface.origin.y, surface.origin.z], DEFAULT_UNITS.SERVER_LENGTH
    )
    axis = UnitVector3D([surface.axis.x, surface.axis.y, surface.axis.z])
    reference = UnitVector3D([surface.reference.x, surface.reference.y, surface.reference.z])

    if surface_type == SurfaceType.SURFACETYPE_CONE:
        result = Cone(origin, surface.radius, surface.half_angle, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_CYLINDER:
        result = Cylinder(origin, surface.radius, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_SPHERE:
        result = Sphere(origin, surface.radius, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_TORUS:
        result = Torus(origin, surface.major_radius, surface.minor_radius, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_PLANE:
        result = PlaneSurface(origin, reference, axis)
    else:
        result = None
    return result


def grpc_curve_to_curve(curve: GRPCCurve) -> Curve:
    """Convert a curve gRPC message to a ``Curve``.

    Parameters
    ----------
    curve : GRPCCurve
        Geometry service gRPC curve message.

    Returns
    -------
    Curve
        Resulting converted curve.
    """
    origin = Point3D([curve.origin.x, curve.origin.y, curve.origin.z])
    try:
        reference = UnitVector3D([curve.reference.x, curve.reference.y, curve.reference.z])
        axis = UnitVector3D([curve.axis.x, curve.axis.y, curve.axis.z])
    except ValueError:
        # curve will be a line
        pass
    if curve.radius != 0:
        result = Circle(origin, curve.radius, reference, axis)
    elif curve.major_radius != 0 and curve.minor_radius != 0:
        result = Ellipse(origin, curve.major_radius, curve.minor_radius, reference, axis)
    elif curve.direction is not None:
        result = Line(
            origin,
            UnitVector3D(
                [
                    curve.direction.x,
                    curve.direction.y,
                    curve.direction.z,
                ]
            ),
        )
    else:
        result = None

    return result


def curve_to_grpc_curve(curve: Curve) -> GRPCCurve:
    """Convert a ``Curve`` object to a curve gRPC message.

    Parameters
    ----------
    curve : Curve
        Curve to convert.

    Returns
    -------
    GRPCCurve
        Return ``Curve`` as a ``ansys.api.geometry.CurveGeometry`` message.
    """
    grpc_curve = None
    origin = point3d_to_grpc_point(curve.origin)

    if isinstance(curve, Line):
        direction = unit_vector_to_grpc_direction(curve.direction)
        grpc_curve = GRPCCurve(origin=origin, direction=direction)
    else:
        reference = unit_vector_to_grpc_direction(curve.dir_x)
        axis = unit_vector_to_grpc_direction(curve.dir_z)

        if isinstance(curve, Circle):
            grpc_curve = GRPCCurve(
                origin=origin, reference=reference, axis=axis, radius=curve.radius.m
            )
        elif isinstance(curve, Ellipse):
            grpc_curve = GRPCCurve(
                origin=origin,
                reference=reference,
                axis=axis,
                major_radius=curve.major_radius.m,
                minor_radius=curve.minor_radius.m,
            )

    return grpc_curve


def trimmed_curve_to_grpc_trimmed_curve(curve: "TrimmedCurve") -> GRPCTrimmedCurve:
    """Convert a ``TrimmedCurve`` to a trimmed curve gRPC message.

    Parameters
    ----------
    curve : TrimmedCurve
        Curve to convert.

    Returns
    -------
    GRPCTrimmedCurve
        Geometry service gRPC ``TrimmedCurve`` message.
    """
    curve_geometry = curve_to_grpc_curve(curve.geometry)
    i_start = curve.interval.start
    i_end = curve.interval.end

    return GRPCTrimmedCurve(
        curve=curve_geometry,
        interval_start=i_start,
        interval_end=i_end,
    )


def line_to_grpc_line(line: Line) -> GRPCLine:
    """Convert a ``Line`` to a line gRPC message.

    Parameters
    ----------
    line : Line
        Line to convert.

    Returns
    -------
    GRPCLine
        Geometry service gRPC ``Line`` message.
    """
    start = line.origin
    end = line.origin + line.direction
    return GRPCLine(start=point3d_to_grpc_point(start), end=point3d_to_grpc_point(end))


def grpc_material_to_material(material: GRPCMaterial) -> Material:
    """Convert a material gRPC message to a ``Material`` class.

    Parameters
    ----------
    material : GRPCMaterial
        Material gRPC message.

    Returns
    -------
    Material
        Converted material.
    """
    properties = []
    density = Quantity(0, UNITS.kg / UNITS.m**3)
    for property in material.material_properties:
        mp = grpc_material_property_to_material_property(property)
        properties.append(mp)
        if mp.type == MaterialPropertyType.DENSITY:
            density = mp.quantity

    return Material(material.name, density, properties)


def grpc_material_property_to_material_property(
    material_property: GRPCMaterialProperty,
) -> MaterialProperty:
    """Convert a material property gRPC message to a ``MaterialProperty`` class.

    Parameters
    ----------
    material_property : GRPCMaterialProperty
        Material property gRPC message.

    Returns
    -------
    MaterialProperty
        Converted material property.
    """
    try:
        mp_type = MaterialPropertyType.from_id(material_property.id)
    except ValueError:
        mp_type = material_property.id

    try:
        mp_quantity = Quantity(material_property.value, material_property.units)
    except (
        UndefinedUnitError,
        TypeError,
    ):
        mp_quantity = material_property.value

    return MaterialProperty(mp_type, material_property.display_name, mp_quantity)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ``BoxUV`` class."""

from enum import Enum

from ansys.geometry.core.shapes.parameterization import Interval, ParamUV
from ansys.geometry.core.typing import Real


class LocationUV(Enum):
    """Provides the enumeration for indicating locations for BoxUV."""

    TopLeft = 1
    TopCenter = 2
    TopRight = 3
    BottomLeft = 4
    BottomCenter = 5
    BottomRight = 6
    LeftCenter = 7
    RightCenter = 8
    Center = 9


class BoxUV:
    """Provides the implementation for ``BoxUV`` class."""

    def __init__(self, range_u: Interval = None, range_v: Interval = None) -> None:
        """Root constructor for BoxUV."""
        if range_u is not None:
            self._interval_u = range_u
        if range_v is not None:
            self._interval_v = range_v

    @property
    def interval_u(self) -> Interval:
        """``u`` interval."""
        return self._interval_u

    @property
    def interval_v(self) -> Interval:
        """``v`` interval."""
        return self._interval_v

    def __eq__(self, other: object) -> bool:
        """Check whether two BoxUV instances are equal."""
        if not isinstance(other, BoxUV):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.interval_u.__eq__(other.interval_u) and self.interval_v.__eq__(other.interval_v)

    def __ne__(self, other: object) -> bool:
        """Check whether two BoxUV instances are not equal."""
        if not isinstance(other, BoxUV):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return not self.interval_u.__eq__(other.interval_u) or not self.interval_v.__eq__(
            other.interval_v
        )

    def is_empty(self):
        """Check if this BoxUV is empty."""
        return self.interval_u.is_empty() or self.interval_v.is_empty()

    def proportion(self, prop_u: Real, prop_v: Real) -> ParamUV:
        """Evaluate the BoxUV at the given proportions."""
        return ParamUV(
            self.interval_u.get_relative_val(prop_u), self.interval_v.get_relative_val(prop_v)
        )

    def get_center(self) -> ParamUV:
        """Evaluate the this BoxUV in the center."""
        return self.proportion(0.5, 0.5)

    def is_negative(self, tolerance_u: Real, tolerance_v: Real) -> bool:
        """Check whether the BoxUV is negative."""
        return self.interval_u.is_negative(tolerance_u) or self.interval_v.is_negative(tolerance_v)

    def contains(self, param: ParamUV) -> bool:
        """Check whether the BoxUV contains a given u and v pair parameter."""
        return self.interval_u.contains(param.u) and self.interval_v.contains(param.v)

    def inflate(self, delta_u: Real, delta_v: Real) -> "BoxUV":
        """Enlarge the BoxUV u and v intervals by deltas."""
        return BoxUV(self.interval_u.inflate(delta_u), self.interval_v.inflate(delta_v))

    def get_corner(self, location: LocationUV) -> ParamUV:
        """Get the corner location of the BoxUV."""
        u = 0
        v = 0
        if (
            location == LocationUV.TopLeft
            or location == LocationUV.BottomLeft
            or location == LocationUV.LeftCenter
        ):
            u = self.interval_u.get_relative_val(0)
        elif (
            location == LocationUV.TopRight
            or location == LocationUV.BottomRight
            or location == LocationUV.RightCenter
        ):
            u = self.interval_u.get_relative_val(1)
        else:
            u = self.interval_u.get_relative_val(0.5)

        if (
            location == LocationUV.BottomRight
            or location == LocationUV.BottomLeft
            or location == LocationUV.BottomCenter
        ):
            v = self.interval_u.get_relative_val(0)
        elif (
            location == LocationUV.TopRight
            or location == LocationUV.TopLeft
            or location == LocationUV.TopCenter
        ):
            v = self.interval_u.get_relative_val(1)
        else:
            v = self.interval_u.get_relative_val(0.5)
        return ParamUV(u, v)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a NURBS curve."""

from functools import cached_property
from typing import TYPE_CHECKING, Optional

from beartype import beartype as check_input_types
import numpy as np
from scipy.integrate import quad

from ansys.geometry.core.math import Matrix44, Point3D
from ansys.geometry.core.math.vector import Vector3D
from ansys.geometry.core.shapes.curves.curve import Curve
from ansys.geometry.core.shapes.curves.curve_evaluation import CurveEvaluation
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
)
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    import geomdl.NURBS as geomdl_nurbs  # noqa: N811


class NURBSCurve(Curve):
    """Represents a NURBS curve.

    Notes
    -----
    This class is a wrapper around the NURBS curve class from the `geomdl` library.
    By leveraging the `geomdl` library, this class provides a high-level interface
    to create and manipulate NURBS curves. The `geomdl` library is a powerful
    library for working with NURBS curves and surfaces. For more information, see
    https://pypi.org/project/geomdl/.

    """

    def __init__(self, geomdl_object: "geomdl_nurbs.Curve" = None):
        """Initialize ``NURBSCurve`` class."""
        try:
            import geomdl.NURBS as geomdl_nurbs  # noqa: N811
        except ImportError as e:  # pragma: no cover
            raise ImportError(
                "The `geomdl` library is required to use the NURBSCurve class. "
                "Please install it using `pip install geomdl`."
            ) from e

        self._nurbs_curve = geomdl_object if geomdl_object else geomdl_nurbs.Curve()

    @property
    def geomdl_nurbs_curve(self) -> "geomdl_nurbs.Curve":
        """Get the underlying NURBS curve.

        Notes
        -----
        This property gives access to the full functionality of the NURBS curve
        coming from the `geomdl` library. Use with caution.
        """
        return self._nurbs_curve

    @property
    def control_points(self) -> list[Point3D]:
        """Get the control points of the curve."""
        return [Point3D(point) for point in self._nurbs_curve.ctrlpts]

    @property
    def degree(self) -> int:
        """Get the degree of the curve."""
        return self._nurbs_curve.degree

    @property
    def knots(self) -> list[Real]:
        """Get the knot vector of the curve."""
        return self._nurbs_curve.knotvector

    @property
    def weights(self) -> list[Real]:
        """Get the weights of the control points."""
        return self._nurbs_curve.weights

    def length(self, num_points: int = None) -> Real:
        """Calculate the length of the NURBS curve.

        Parameters
        ----------
        num_points : int, default: None
            Number of points to sample along the curve for length calculation.

        Returns
        -------
        Real
            Length of the NURBS curve.
        """
        if num_points is None:
            num_spans = len(self._nurbs_curve.knotvector) - (2 * self._nurbs_curve.degree) - 1
            num_points = max(
                num_spans * 10, 50
            )  # 10 samples per span, floor of 50 ensures accuracy

        self._nurbs_curve.sample_size = num_points

        def arc_length_func(u):
            deriv = self._nurbs_curve.derivatives(u, order=1)[1]
            return np.linalg.norm(deriv)

        # Integrate over the curve's domain
        start_u = self._nurbs_curve.knotvector[self._nurbs_curve.degree]
        end_u = self._nurbs_curve.knotvector[-(self._nurbs_curve.degree + 1)]

        length, _ = quad(arc_length_func, start_u, end_u)
        return length

    @classmethod
    @check_input_types
    def from_control_points(
        cls,
        control_points: list[Point3D],
        degree: int,
        knots: list[Real],
        weights: list[Real] = None,
    ) -> "NURBSCurve":
        """Create a NURBS curve from control points.

        Parameters
        ----------
        control_points : list[Point3D]
            Control points of the curve.
        degree : int
            Degree of the curve.
        knots : list[Real]
            Knot vector of the curve.
        weights : list[Real], optional
            Weights of the control points.

        Returns
        -------
        NURBSCurve
            NURBS curve.

        """
        curve = cls()
        curve._nurbs_curve.degree = degree
        curve._nurbs_curve.ctrlpts = control_points
        curve._nurbs_curve.knotvector = knots
        if weights:
            curve._nurbs_curve.weights = weights

        # Verify the curve is valid
        try:
            curve._nurbs_curve._check_variables()
        except ValueError as e:
            raise ValueError(f"Invalid NURBS curve: {e}")

        return curve

    @classmethod
    @check_input_types
    def fit_curve_from_points(
        cls,
        points: list[Point3D],
        degree: int,
    ) -> "NURBSCurve":
        """Fit a NURBS curve to a set of points.

        Parameters
        ----------
        points : list[Point3D]
            Points to fit the curve to.
        degree : int
            Degree of the curve.

        Returns
        -------
        NURBSCurve
            Fitted NURBS curve.

        """
        from geomdl import fitting

        # Convert points to a format suitable for the fitting function
        converted_points = [[*pt] for pt in points]

        # Fit the curve to the points
        curve = fitting.interpolate_curve(converted_points, degree)

        # Construct the NURBSCurve object
        nurbs_curve = cls()
        nurbs_curve._nurbs_curve.degree = curve.degree
        nurbs_curve._nurbs_curve.ctrlpts = [Point3D(entry) for entry in curve.ctrlpts]
        nurbs_curve._nurbs_curve.knotvector = curve.knotvector
        nurbs_curve._nurbs_curve.weights = curve.weights

        # Verify the curve is valid
        try:
            nurbs_curve._nurbs_curve._check_variables()
        except ValueError as e:
            raise ValueError(f"Invalid NURBS curve: {e}")

        return nurbs_curve

    def __eq__(self, other: "NURBSCurve") -> bool:
        """Determine if two curves are equal."""
        if not isinstance(other, NURBSCurve):
            return False
        return (
            self._nurbs_curve.degree == other._nurbs_curve.degree
            and self._nurbs_curve.ctrlpts == other._nurbs_curve.ctrlpts
            and self._nurbs_curve.knotvector == other._nurbs_curve.knotvector
            and self._nurbs_curve.weights == other._nurbs_curve.weights
        )

    def parameterization(self) -> Parameterization:
        """Get the parametrization of the NURBS curve.

        The parameter is defined in the interval [0, 1] by default. Information
        is provided about the parameter type and form.

        Returns
        -------
        Parameterization
            Information about how the NURBS curve is parameterized.
        """
        return Parameterization(
            ParamForm.OTHER,
            ParamType.OTHER,
            Interval(start=self._nurbs_curve.domain[0], end=self._nurbs_curve.domain[1]),
        )

    def transformed_copy(self, matrix: Matrix44) -> "NURBSCurve":
        """Create a transformed copy of the curve.

        Parameters
        ----------
        matrix : Matrix44
            Transformation matrix.

        Returns
        -------
        NURBSCurve
            Transformed copy of the curve.
        """
        control_points = []
        for point in self._nurbs_curve.ctrlpts:
            # Transform the control point using the transformation matrix
            transformed_point = matrix @ np.array([*point, 1])
            control_points.append(Point3D(transformed_point[:3]))

        return NURBSCurve.from_control_points(
            control_points,
            self._nurbs_curve.degree,
            self._nurbs_curve.knotvector,
            self._nurbs_curve.weights,
        )

    def evaluate(self, parameter: Real) -> CurveEvaluation:
        """Evaluate the curve at the given parameter.

        Parameters
        ----------
        parameter : Real
            Parameter to evaluate the curve at.

        Returns
        -------
        CurveEvaluation
            Evaluation of the curve at the given parameter.
        """
        return NURBSCurveEvaluation(self, parameter)

    def contains_param(self, param: Real) -> bool:  # noqa: D102
        raise NotImplementedError("contains_param() is not implemented.")

    def contains_point(self, point: Point3D) -> bool:  # noqa: D102
        raise NotImplementedError("contains_point() is not implemented.")

    def project_point(
        self, point: Point3D, initial_guess: Optional[Real] = None
    ) -> CurveEvaluation:
        """Project a point to the NURBS curve.

        This method returns the evaluation at the closest point.

        Parameters
        ----------
        point : Point3D
            Point to project to the curve.
        initial_guess : Real, optional
            Initial guess for the optimization algorithm. If not provided, the midpoint
            of the domain is used.

        Returns
        -------
        CurveEvaluation
            Evaluation at the closest point on the curve.

        Notes
        -----
        Based on `the NURBS book <https://link.springer.com/book/10.1007/978-3-642-59223-2>`_,
        the projection of a point to a NURBS curve is the solution to the following optimization
        problem: minimize the distance between the point and the curve. The distance is defined
        as the Euclidean distance squared. For more information, please refer to
        the implementation of the `distance_squared` function.
        """
        import numpy as np
        from scipy.optimize import minimize

        # Function to minimize (distance squared)
        def distance_squared(
            u: float, geomdl_nurbs_curbe: "geomdl_nurbs.Curve", point: np.ndarray
        ) -> np.ndarray:
            point_on_curve = np.array(geomdl_nurbs_curbe.evaluate_single(u))
            return np.sum((point_on_curve - point) ** 2)

        # Define the domain and initial guess (midpoint of the domain by default)
        domain = self._nurbs_curve.domain
        initial_guess = initial_guess if initial_guess else (domain[0] + domain[1]) / 2

        # Minimize the distance squared
        result = minimize(
            distance_squared,
            initial_guess,
            bounds=[domain],
            args=(self._nurbs_curve, np.array(point)),
        )

        # Closest point on the curve
        u_min = result.x[0]

        # Return the evaluation at the closest point
        return self.evaluate(u_min)


class NURBSCurveEvaluation(CurveEvaluation):
    """Provides evaluation of a NURBS curve at a given parameter.

    Parameters
    ----------
    nurbs_curve: ~ansys.geometry.core.shapes.curves.nurbs.NURBSCurve
        NURBS curve to evaluate.
    parameter: Real
        Parameter to evaluate the NURBS curve at.
    """

    def __init__(self, nurbs_curve: NURBSCurve, parameter: Real) -> None:
        """Initialize the ``NURBSCurveEvaluation`` class."""
        self._parameter = parameter
        self._point_eval, self._first_deriv_eval, self._second_deriv_eval = (
            nurbs_curve.geomdl_nurbs_curve.derivatives(parameter, 2)
        )

    @property
    def parameter(self) -> Real:
        """Parameter that the evaluation is based upon."""
        return self._parameter

    @cached_property
    def position(self) -> Point3D:
        """Position of the evaluation."""
        return Point3D(self._point_eval)

    @cached_property
    def first_derivative(self) -> Vector3D:
        """First derivative of the evaluation."""
        return Vector3D(self._first_deriv_eval)

    @cached_property
    def second_derivative(self) -> Vector3D:
        """Second derivative of the evaluation."""
        return Vector3D(self._second_deriv_eval)

    @cached_property
    def curvature(self) -> Real:
        """Curvature of the evaluation."""
        # For a curve, the curvature is the magnitude of the cross product
        # of the first and second derivatives divided by the cube of the
        # magnitude of the first derivative. For more information, please refer
        # to https://en.wikipedia.org/wiki/Curvature#General_expressions.
        return (
            self.first_derivative.cross(self.second_derivative).magnitude
            / self.first_derivative.magnitude**3
        )



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing default connection parameters."""

import os
from pathlib import Path

DEFAULT_HOST = os.environ.get("ANSRV_GEO_HOST", "127.0.0.1")
"""Default for the HOST name.

By default, PyAnsys Geometry searches for the environment variable ``ANSRV_GEO_HOST``,
and if this variable does not exist, PyAnsys Geometry uses ``127.0.0.1`` as the host.
"""

DEFAULT_PORT: int = int(os.environ.get("ANSRV_GEO_PORT", 50051))
"""Default for the HOST port.

By default, PyAnsys Geometry searches for the environment variable ``ANSRV_GEO_PORT``,
and if this variable does not exist, PyAnsys Geometry uses ``50051`` as the port.
"""

MAX_MESSAGE_LENGTH = int(os.environ.get("PYGEOMETRY_MAX_MESSAGE_LENGTH", 256 * 1024**2))
"""Default for the gRPC maximum message length.

By default, PyAnsys Geometry searches for the environment variable
``PYGEOMETRY_MAX_MESSAGE_LENGTH``, and if this variable does not exist,
it uses ``256Mb`` as the maximum message length.
"""

GEOMETRY_SERVICE_DOCKER_IMAGE = "ghcr.io/ansys/geometry"
"""Default for the Geometry service Docker image location.

Tag is dependent on what OS service is requested.
"""

DEFAULT_PIM_CONFIG = Path(__file__).parent / "pim_configuration.json"
"""Default for the PIM configuration when running PIM Light.

This parameter is only to be used when PIM Light is being run.
"""



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides tools for preparing geometry for use with simulation."""

from typing import TYPE_CHECKING

from beartype import beartype as check_input_types

from ansys.geometry.core.connection import GrpcClient
from ansys.geometry.core.connection.backend import BackendType
from ansys.geometry.core.errors import GeometryRuntimeError
from ansys.geometry.core.logger import LOG
from ansys.geometry.core.misc.auxiliary import (
    get_bodies_from_ids,
    get_design_from_edge,
    get_design_from_face,
)
from ansys.geometry.core.misc.checks import check_type_all_elements_in_iterable, min_backend_version
from ansys.geometry.core.tools.problem_areas import LogoProblemArea
from ansys.geometry.core.tools.repair_tool_message import RepairToolMessage
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face


class PrepareTools:
    """Prepare tools for PyAnsys Geometry.

    Parameters
    ----------
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    _internal_use : bool, optional
        Internal flag to prevent direct instantiation by users.
        This parameter is for internal use only.

    Raises
    ------
    GeometryRuntimeError
        If the class is instantiated directly by users instead of through the modeler.

    Notes
    -----
    This class should not be instantiated directly. Use
    ``modeler.prepare_tools`` instead.
    """

    def __init__(self, grpc_client: GrpcClient, _internal_use: bool = False):
        """Initialize Prepare Tools class."""
        if not _internal_use:
            raise GeometryRuntimeError(
                "PrepareTools should not be instantiated directly. "
                "Use 'modeler.prepare_tools' to access prepare tools."
            )
        self._grpc_client = grpc_client

    @min_backend_version(25, 1, 0)
    def extract_volume_from_faces(
        self, sealing_faces: list["Face"], inside_faces: list["Face"]
    ) -> list["Body"]:
        """Extract a volume from input faces.

        Creates a volume (typically a flow volume) from a list of faces that seal the volume
        and one or more faces that define the wetted surface (inside faces of the solid).

        Parameters
        ----------
        sealing_faces : list[Face]
            List of faces that seal the volume.
        inside_faces : list[Face]
            List of faces that define the interior of the solid.

        Returns
        -------
        list[Body]
            List of created bodies.

        Warnings
        --------
        This method is only available starting on Ansys release 25R1.
        """
        from ansys.geometry.core.designer.face import Face

        if not sealing_faces or not inside_faces:
            self._grpc_client.log.info("No sealing faces or inside faces provided...")
            return []

        # Verify inputs
        check_type_all_elements_in_iterable(sealing_faces, Face)
        check_type_all_elements_in_iterable(inside_faces, Face)

        parent_design = get_design_from_face(sealing_faces[0])

        response = self._grpc_client._services.prepare_tools.extract_volume_from_faces(
            sealing_faces=sealing_faces,
            inside_faces=inside_faces,
        )

        if response.get("success"):
            bodies_ids = response.get("created_bodies")
            if len(bodies_ids) > 0:
                parent_design._update_design_inplace()
            return get_bodies_from_ids(parent_design, bodies_ids)
        else:
            self._grpc_client.log.info("Failed to extract volume from faces...")
            return []

    @min_backend_version(25, 1, 0)
    def extract_volume_from_edge_loops(
        self, sealing_edges: list["Edge"], inside_faces: list["Face"] = None
    ) -> list["Body"]:
        """Extract a volume from input edge loops.

        Creates a volume (typically a flow volume) from a list of edge loops that seal the volume.
        and one or more faces that define the wetted surface (inside faces of the solid).

        Parameters
        ----------
        sealing_edges : list[Edge]
            List of faces that seal the volume.
        inside_faces : list[Face], optional
            List of faces that define the interior of the solid (not always necessary).

        Returns
        -------
        list[Body]
            List of created bodies.

        Warnings
        --------
        This method is only available starting on Ansys release 25R1.
        """
        from ansys.geometry.core.designer.edge import Edge
        from ansys.geometry.core.designer.face import Face

        if not sealing_edges:
            self._grpc_client.log.info("No sealing edges provided...")
            return []

        # Assign default values to inside_faces
        inside_faces = [] if inside_faces is None else inside_faces

        # Verify inputs
        check_type_all_elements_in_iterable(sealing_edges, Edge)
        check_type_all_elements_in_iterable(inside_faces, Face)

        parent_design = get_design_from_edge(sealing_edges[0])

        response = self._grpc_client._services.prepare_tools.extract_volume_from_edge_loops(
            sealing_edges=sealing_edges,
            inside_faces=inside_faces,
        )

        if response.get("success"):
            bodies_ids = response.get("created_bodies")
            if len(bodies_ids) > 0:
                parent_design._update_design_inplace()
            return get_bodies_from_ids(parent_design, bodies_ids)
        else:
            self._grpc_client.log.info("Failed to extract volume from edge loops...")
            return []

    def remove_rounds(self, faces: list["Face"], auto_shrink: bool = False) -> bool:
        """Remove rounds from geometry.

        Tries to remove rounds from geometry. Faces to be removed are input to the method.

        Parameters
        ----------
        round_faces : list[Face]
            List of rounds faces to be removed
        auto_shrink : bool, default: False
            Whether to shrink the geometry after removing rounds. Fills in the gaps
            left by the removed rounds.

        Returns
        -------
        bool
            ``True`` if successful, ``False`` if failed.
        """
        from ansys.geometry.core.designer.face import Face

        if not faces:
            self._grpc_client.log.info("No faces provided...")
            return []

        # Verify inputs
        check_type_all_elements_in_iterable(faces, Face)

        parent_design = get_design_from_face(faces[0])
        response = self._grpc_client._services.prepare_tools.remove_rounds(
            rounds=faces,
            auto_shrink=auto_shrink,
        )

        if response.get("success"):
            parent_design._update_design_inplace()
        else:
            self._grpc_client.log.info("Failed to remove rounds...")

        return response.get("success")

    @min_backend_version(24, 2, 0)
    def share_topology(
        self, bodies: list["Body"], tol: Real = 0.0, preserve_instances: bool = False
    ) -> bool:
        """Share topology between the chosen bodies.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies to share topology between.
        tol : Real
            Maximum distance between bodies.
        preserve_instances : bool
            Whether instances are preserved.

        Returns
        -------
        bool
            ``True`` if successful, ``False`` if failed.

        Warnings
        --------
        This method is only available starting on Ansys release 24R2.
        """
        from ansys.geometry.core.designer.body import Body

        if not bodies:
            return False

        # Verify inputs
        check_type_all_elements_in_iterable(bodies, Body)

        response = self._grpc_client._services.prepare_tools.share_topology(
            bodies=bodies,
            tolerance=tol,
            preserve_instances=preserve_instances,
        )

        return response.get("success")

    @min_backend_version(25, 2, 0)
    def enhanced_share_topology(
        self, bodies: list["Body"], tol: Real = 0.0, preserve_instances: bool = False
    ) -> RepairToolMessage:
        """Share topology between the chosen bodies.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies to share topology between.
        tol : Real
            Maximum distance between bodies.
        preserve_instances : bool
            Whether instances are preserved.

        Returns
        -------
        RepairToolMessage
            Message containing number of problem areas found/fixed, created and/or modified bodies.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.body import Body

        if not bodies:
            return RepairToolMessage(
                success=False, created_bodies=[], modified_bodies=[], found=0, repaired=0
            )

        # Verify inputs
        check_type_all_elements_in_iterable(bodies, Body)

        response = self._grpc_client._services.prepare_tools.enhanced_share_topology(
            bodies=bodies,
            tolerance=tol,
            preserve_instances=preserve_instances,
        )

        message = RepairToolMessage(
            success=response.get("success"),
            created_bodies=response.get("created_bodies_monikers"),
            modified_bodies=response.get("modified_bodies_monikers"),
            found=response.get("found"),
            repaired=response.get("repaired"),
        )
        return message

    @check_input_types
    @min_backend_version(25, 2, 0)
    def find_logos(
        self, bodies: list["Body"] = None, min_height: Real = None, max_height: Real = None
    ) -> "LogoProblemArea":
        """Detect logos in geometry.

        Detects logos, using a list of bodies if provided.
        The logos are returned as a list of faces.

        Parameters
        ----------
        bodies : list[Body], optional
            List of bodies where logos should be detected
        min_height : real, optional
            The minimum height when searching for logos
        max_height: real, optional
            The minimum height when searching for logos

        Returns
        -------
        LogoProblemArea
            Problem area with logo faces.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.body import Body

        if BackendType.is_linux_service(
            self._grpc_client.backend_type
        ) and self._grpc_client.backend_version < (26, 1, 0):
            # not yet available on Linux until 26.1.0
            LOG.warning("Logo detection not available on Linux")
            return

        # Verify inputs
        if bodies and len(bodies) > 0:
            check_type_all_elements_in_iterable(bodies, Body)

        bodies = [] if bodies is None else bodies
        response = self._grpc_client._services.prepare_tools.find_logos(
            bodies=bodies,
            min_height=min_height,
            max_height=max_height,
        )

        return LogoProblemArea(
            id=response.get("id"),
            grpc_client=self._grpc_client,
            face_ids=response.get("face_ids"),
        )

    @check_input_types
    @min_backend_version(25, 2, 0)
    def find_and_remove_logos(
        self, bodies: list["Body"] = None, min_height: Real = None, max_height: Real = None
    ) -> bool:
        """Detect and remove logos in geometry.

        Detects and remove logos, using a list of bodies if provided.

        Parameters
        ----------
        bodies : list[Body], optional
            List of bodies where logos should be detected and removed.
        min_height : real, optional
            The minimum height when searching for logos
        max_height: real, optional
            The minimum height when searching for logos

        Returns
        -------
        Boolean value indicating whether the operation was successful.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.body import Body

        if BackendType.is_linux_service(
            self._grpc_client.backend_type
        ) and self._grpc_client.backend_version < (26, 1, 0):
            # not yet available on Linux until 26.1.0
            LOG.warning("Logo detection not available on Linux")
            return

        # Verify inputs
        if bodies and len(bodies) > 0:
            check_type_all_elements_in_iterable(bodies, Body)

        bodies = [] if bodies is None else bodies
        response = self._grpc_client._services.prepare_tools.find_and_remove_logos(
            bodies=bodies,
            min_height=min_height,
            max_height=max_height,
        )

        return response.get("success")



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing an ellipse."""

from typing import TYPE_CHECKING

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity
from scipy.spatial.transform import Rotation as SpatialRotation

from ansys.geometry.core.math.matrix import Matrix33, Matrix44
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point2D, Point3D
from ansys.geometry.core.math.vector import Vector3D
from ansys.geometry.core.misc.checks import graphics_required
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.shapes.curves.ellipse import Ellipse
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv


class SketchEllipse(SketchFace, Ellipse):
    """Provides for modeling an ellipse.

    Parameters
    ----------
    center: Point2D
        Center point of the ellipse.
    major_radius : ~pint.Quantity | Distance | Real
        Major radius of the ellipse.
    minor_radius : ~pint.Quantity | Distance | Real
        Minor radius of the ellipse.
    angle : ~pint.Quantity | Angle | Real, default: 0
        Placement angle for orientation alignment.
    plane : Plane, optional
        Plane containing the sketched ellipse, which is the global XY plane
        by default.
    """

    @check_input_types
    def __init__(
        self,
        center: Point2D,
        major_radius: Quantity | Distance | Real,
        minor_radius: Quantity | Distance | Real,
        angle: Quantity | Angle | Real = 0,
        plane: Plane = Plane(),
    ):
        """Initialize the ellipse."""
        # Call SketchFace init method
        SketchFace.__init__(self)

        # Store the 2D center of the ellipse
        self._center = center

        self._major_radius = (
            major_radius if isinstance(major_radius, Distance) else Distance(major_radius)
        )
        self._minor_radius = (
            minor_radius if isinstance(minor_radius, Distance) else Distance(minor_radius)
        )

        self._angle_offset = angle if isinstance(angle, Angle) else Angle(angle)

        # Call Ellipse init method
        self._init_primitive_ellipse_from_plane(
            plane, self._major_radius, self._minor_radius, self._angle_offset
        )

    def _init_primitive_ellipse_from_plane(
        self,
        plane: Plane,
        major_radius: Distance | None = None,
        minor_radius: Distance | None = None,
        angle: Angle | None = None,
    ) -> None:
        """Initialize correctly the underlying primitive ``Ellipse`` class.

        Parameters
        ----------
        plane : Plane
            Plane containing the sketched ellipse.
        major_radius : Distance, default: None
            Major radius of the ellipse (if any).
        minor_radius : Distance, default: None
            Minor radius of the ellipse (if any).
        angle : Angle, default: None
            Placement angle for orientation alignment.
        """
        major_radius = major_radius if major_radius else self.major_radius
        minor_radius = minor_radius if minor_radius else self.minor_radius
        angle = angle if angle else self.angle

        # Call Ellipse init method
        center_global = plane.origin + Point3D(
            self.center[0] * plane.direction_x + self.center[1] * plane.direction_y,
            unit=self.center.base_unit,
        )

        angle_rad = angle.value.m_as(UNITS.radian)
        new_rotated_dir_x = Vector3D(
            [
                np.cos(angle_rad) * plane.direction_x.x - np.sin(angle_rad) * plane.direction_x.y,
                np.sin(angle_rad) * plane.direction_x.x + np.cos(angle_rad) * plane.direction_x.y,
                plane.direction_x.z,
            ]
        )

        Ellipse.__init__(
            self, center_global, major_radius, minor_radius, new_rotated_dir_x, plane.direction_z
        )

    @property
    def center(self) -> Point2D:
        """Center point of the ellipse."""
        return self._center

    @property
    def angle(self) -> Quantity:
        """Orientation angle of the ellipse."""
        return self._angle_offset.value

    @property
    def perimeter(self) -> Quantity:
        """Perimeter of the circle.

        Notes
        -----
        This property resolves the dilemma between using the ``SkethFace.perimeter``
        property and the ``Ellipse.perimeter`` property.
        """
        return Ellipse.perimeter.fget(self)

    @property
    @graphics_required
    def visualization_polydata(self) -> "pv.PolyData":
        """VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        import pyvista as pv

        rotation = Matrix33(
            SpatialRotation.from_euler(
                "xyz", [0, 0, self._angle_offset.value.m_as(UNITS.radian)], degrees=False
            ).as_matrix()
        )

        transformation_matrix = Matrix44(
            [
                [
                    rotation[0, 0],
                    rotation[0, 1],
                    rotation[0, 2],
                    self.center.x.m_as(DEFAULT_UNITS.LENGTH),
                ],
                [
                    rotation[1, 0],
                    rotation[1, 1],
                    rotation[1, 2],
                    self.center.y.m_as(DEFAULT_UNITS.LENGTH),
                ],
                [
                    rotation[2, 0],
                    rotation[2, 1],
                    rotation[2, 2],
                    0,
                ],
                [0, 0, 0, 1],
            ]
        )

        return pv.Ellipse(
            semi_major_axis=self.major_radius.m_as(DEFAULT_UNITS.LENGTH),
            semi_minor_axis=self.minor_radius.m_as(DEFAULT_UNITS.LENGTH),
        ).transform(transformation_matrix, inplace=True)

    def plane_change(self, plane: Plane) -> None:
        """Redefine the plane containing ``SketchEllipse`` objects.

        Parameters
        ----------
        plane : Plane
            Desired new plane that is to contain the sketched ellipse.

        Notes
        -----
        This implies that their 3D definition might suffer changes.
        """
        # Reinitialize the Circle definition for the given plane
        self._init_primitive_ellipse_from_plane(plane)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ``TrimmedSurface`` class."""

from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.shapes.box_uv import BoxUV
from ansys.geometry.core.shapes.parameterization import ParamUV
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.surface_evaluation import SurfaceEvaluation
from ansys.geometry.core.typing import Real


class TrimmedSurface:
    """Represents a trimmed surface.

    A trimmed surface is a surface that has a boundary. This boundary comes in the form of a
    bounding BoxUV.

    Parameters
    ----------
    face : Face
        Face that the trimmed surface belongs to.
    geometry : Surface
        Underlying mathematical representation of the surface.
    """

    def __init__(self, geometry: Surface, box_uv: BoxUV):
        """Initialize an instance of a trimmed surface."""
        self._geometry = geometry
        self._box_uv = box_uv

    @property
    def geometry(self) -> Surface:
        """Underlying mathematical surface."""
        return self._geometry

    @property
    def box_uv(self) -> BoxUV:
        """Bounding BoxUV of the surface."""
        return self._box_uv

    def get_proportional_parameters(self, param_uv: ParamUV) -> ParamUV:
        """Convert non-proportional parameters into proportional parameters.

        Parameters
        ----------
        param_uv : ParamUV
            Non-proportional UV parameters.

        Returns
        -------
        ParamUV
            Proportional (from 0-1) UV parameters.
        """
        bounds_u = self.box_uv.interval_u
        bounds_v = self.box_uv.interval_v
        u = param_uv.u
        v = param_uv.v
        return (
            (u - bounds_u.start) / bounds_u.get_span(),
            (v - bounds_v.start) / bounds_v.get_span(),
        )

    def normal(self, u: Real, v: Real) -> UnitVector3D:
        """Provide the normal to the surface.

        Parameters
        ----------
        u : Real
            First coordinate of the 2D representation of a surface in UV space.
        v : Real
            Second coordinate of the 2D representation of a surface in UV space.

        Returns
        -------
        UnitVector3D
            Unit vector is normal to the surface at the given UV coordinates.
        """
        return self.evaluate_proportion(u, v).normal

    def project_point(self, point: Point3D) -> SurfaceEvaluation:
        """Project a point onto the surface and evaluate it at that location.

        Parameters
        ----------
        point : Point3D
            Point to project onto the surface.

        Returns
        -------
        SurfaceEvaluation
            Resulting evaluation.
        """
        return self.geometry.project_point(point)

    def evaluate_proportion(self, u: Real, v: Real) -> SurfaceEvaluation:
        """Evaluate the surface at proportional u and v parameters.

        Parameters
        ----------
        u : Real
            U parameter in the proportional range [0,1].
        v : Real
            V parameter in the proportional range [0,1].

        Returns
        -------
        SurfaceEvaluation
            Resulting surface evaluation.
        """
        bounds_u = self.box_uv.interval_u
        bounds_v = self.box_uv.interval_v
        return self.geometry.evaluate(
            ParamUV(
                bounds_u.start + bounds_u.get_span() * u,
                bounds_v.start + bounds_v.get_span() * v,
            )
        )

    # TODO: perimeter, area?
    # https://github.com/ansys/pyansys-geometry/issues/1319


class ReversedTrimmedSurface(TrimmedSurface):
    """Represents a reversed trimmed surface.

    When a surface is reversed, its normal vector is negated to provide the proper
    outward facing vector.

    Parameters
    ----------
    face : Face
        Face that the trimmed surface belongs to.
    geometry : Surface
        Underlying mathematical representation of the surface.
    """

    def __init__(self, geometry: Surface, box_uv: BoxUV):
        """Initialize an instance of a reversed trimmed surface."""
        super().__init__(geometry, box_uv)

    def normal(self, u: Real, v: Real) -> UnitVector3D:  # noqa: D102
        return -self.evaluate_proportion(u, v).normal

    def project_point(self, point: Point3D) -> SurfaceEvaluation:  # noqa: D102
        evaluation = self.geometry.project_point(point)
        evaluation.normal = -evaluation.normal
        return evaluation



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for evaluating a surface."""

from functools import cached_property

from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.shapes.parameterization import ParamUV
from ansys.geometry.core.typing import Real


class SurfaceEvaluation:
    """Provides for evaluating a surface."""

    def __init__(self, parameter: ParamUV) -> None:
        """Initialize the ``SurfaceEvaluation`` class."""
        self._parameter = parameter

    @property
    def parameter(self) -> ParamUV:
        """Parameter that the evaluation is based upon."""
        raise NotImplementedError("Each evaluation must provide the parameter definition.")

    @cached_property
    def position(self) -> Point3D:
        """Point on the surface, based on the evaluation."""
        raise NotImplementedError("Each evaluation must provide the position definition.")

    @cached_property
    def normal(self) -> UnitVector3D:
        """Normal to the surface."""
        raise NotImplementedError("Each evaluation must provide the position definition.")

    @cached_property
    def u_derivative(self) -> Vector3D:
        """First derivative with respect to the U parameter."""
        raise NotImplementedError("Each evaluation must provide the u-derivative definition.")

    @cached_property
    def v_derivative(self) -> Vector3D:
        """First derivative with respect to the V parameter."""
        raise NotImplementedError("Each evaluation must provide the v-derivative definition.")

    @cached_property
    def uu_derivative(self) -> Vector3D:
        """Second derivative with respect to the U parameter."""
        raise NotImplementedError("Each evaluation must provide the uu-derivative definition.")

    @cached_property
    def uv_derivative(self) -> Vector3D:
        """The second derivative with respect to the U and V parameters."""
        raise NotImplementedError("Each evaluation must provide the uv-derivative definition.")

    @cached_property
    def vv_derivative(self) -> Vector3D:
        """The second derivative with respect to v."""
        raise NotImplementedError("Each evaluation must provide the vv-derivative definition.")

    @cached_property
    def min_curvature(self) -> Real:
        """Minimum curvature."""
        raise NotImplementedError("Each evaluation must provide the minimum curvature definition.")

    @cached_property
    def min_curvature_direction(self) -> UnitVector3D:
        """Minimum curvature direction."""
        raise NotImplementedError(
            "Each evaluation must provide the minimum curvature direction definition."
        )

    @cached_property
    def max_curvature(self) -> Real:
        """Maximum curvature."""
        raise NotImplementedError("Each evaluation must provide the maximum curvature definition.")

    @cached_property
    def max_curvature_direction(self) -> UnitVector3D:
        """Maximum curvature direction."""
        raise NotImplementedError(
            "Each evaluation must provide the maximum curvature direction definition."
        )



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the faces service implementation (abstraction layer)."""

from abc import ABC, abstractmethod

import grpc


class GRPCFacesService(ABC):  # pragma: no cover
    """Faces service for gRPC communication with the Geometry server.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    def __init__(self, channel: grpc.Channel):
        """Initialize the GRPCFacesService class."""
        pass

    @abstractmethod
    def get_surface(self, **kwargs) -> dict:
        """Get the surface of a face."""
        pass

    @abstractmethod
    def get_box_uv(self, **kwargs) -> dict:
        """Get the UV box of a face."""
        pass

    @abstractmethod
    def get_area(self, **kwargs) -> dict:
        """Get the area of a face."""
        pass

    @abstractmethod
    def get_edges(self, **kwargs) -> dict:
        """Get the edges of a face."""
        pass

    @abstractmethod
    def get_vertices(self, **kwargs) -> dict:
        """Get the vertices of a face."""
        pass

    @abstractmethod
    def get_loops(self, **kwargs) -> dict:
        """Get the loops of a face."""
        pass

    @abstractmethod
    def get_color(self, **kwargs) -> dict:
        """Get the color of a face."""
        pass

    @abstractmethod
    def get_bounding_box(self, **kwargs) -> dict:
        """Get the bounding box of a face."""
        pass

    @abstractmethod
    def set_color(self, **kwargs) -> dict:
        """Set the color of a face."""
        pass

    @abstractmethod
    def get_normal(self, **kwargs) -> dict:
        """Get the normal of a face."""
        pass

    @abstractmethod
    def evaluate(self, **kwargs) -> dict:
        """Evaluate a face at a given parameter."""
        pass

    @abstractmethod
    def create_iso_parametric_curve(self, **kwargs) -> dict:
        """Create an iso-parametric curve on a face."""
        pass



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the bodies service implementation for v0."""

import grpc
import pint

from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

from ..base.bodies import GRPCBodyService
from ..base.conversions import from_measurement_to_server_angle, from_measurement_to_server_length
from .conversions import (
    _nurbs_curves_compatibility,
    build_grpc_id,
    from_frame_to_grpc_frame,
    from_grpc_material_to_material,
    from_grpc_point_to_point3d,
    from_grpc_tess_to_pd,
    from_plane_to_grpc_plane,
    from_point3d_to_grpc_point,
    from_sketch_shapes_to_grpc_geometries,
    from_tess_options_to_grpc_tess_options,
    from_trimmed_curve_to_grpc_trimmed_curve,
    from_trimmed_surface_to_grpc_trimmed_surface,
    from_unit_vector_to_grpc_direction,
)


class GRPCBodyServiceV0(GRPCBodyService):
    """Body service for gRPC communication with the Geometry server.

    This class provides methods to create and manipulate bodies in the
    Geometry server using gRPC. It is specifically designed for the v0
    version of the Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2_grpc import BodiesStub

        self.stub = BodiesStub(channel)

    @protect_grpc
    def create_sphere_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreateSphereBodyRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSphereBodyRequest(
            name=kwargs["name"],
            parent=kwargs["parent"],
            center=from_point3d_to_grpc_point(kwargs["center"]),
            radius=from_measurement_to_server_length(kwargs["radius"]),
        )

        # Call the gRPC service
        resp = self.stub.CreateSphereBody(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def create_extruded_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreateExtrudedBodyRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateExtrudedBodyRequest(
            name=kwargs["name"],
            parent=kwargs["parent_id"],
            plane=from_plane_to_grpc_plane(kwargs["sketch"].plane),
            distance=from_measurement_to_server_length(kwargs["distance"]) * kwargs["direction"],
            geometries=from_sketch_shapes_to_grpc_geometries(
                kwargs["sketch"].plane, kwargs["sketch"].edges, kwargs["sketch"].faces
            ),
        )

        # HACK: we should inform the user that NURBS curve sketches are not supported
        # prior to 26R1... and if passed, raise an error
        _nurbs_curves_compatibility(kwargs["backend_version"], request.geometries)

        # Call the gRPC service
        resp = self.stub.CreateExtrudedBody(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def create_sweeping_profile_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreateSweepingProfileRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSweepingProfileRequest(
            name=kwargs["name"],
            parent=kwargs["parent_id"],
            plane=from_plane_to_grpc_plane(kwargs["sketch"].plane),
            geometries=from_sketch_shapes_to_grpc_geometries(
                kwargs["sketch"].plane, kwargs["sketch"].edges, kwargs["sketch"].faces
            ),
            path=[from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in kwargs["path"]],
        )

        # HACK: we should inform the user that NURBS curve sketches are not supported
        # prior to 26R1... and if passed, raise an error
        _nurbs_curves_compatibility(kwargs["backend_version"], request.geometries)

        # Call the gRPC service
        resp = self.stub.CreateSweepingProfile(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def create_sweeping_chain(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreateSweepingChainRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSweepingChainRequest(
            name=kwargs["name"],
            parent=kwargs["parent_id"],
            path=[from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in kwargs["path"]],
            chain=[from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in kwargs["chain"]],
        )

        # Call the gRPC service
        resp = self.stub.CreateSweepingChain(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def sweep_with_guide(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
        from ansys.api.geometry.v0.bodies_pb2 import (
            SweepWithGuideRequest,
            SweepWithGuideRequestData,
        )

        # Create request object - assumes all inputs are valid and of the proper type
        request = SweepWithGuideRequest(
            request_data=[
                SweepWithGuideRequestData(
                    name=data.name,
                    parent=EntityIdentifier(id=data.parent_id),
                    plane=from_plane_to_grpc_plane(data.sketch.plane),
                    geometries=from_sketch_shapes_to_grpc_geometries(
                        data.sketch.plane, data.sketch.edges, data.sketch.faces
                    ),
                    path=from_trimmed_curve_to_grpc_trimmed_curve(data.path),
                    guide=from_trimmed_curve_to_grpc_trimmed_curve(data.guide),
                    tight_tolerance=data.tight_tolerance,
                )
                for data in kwargs["sweep_data"]
            ],
        )

        # Call the gRPC service
        resp = self.stub.SweepWithGuide(request=request)

        # Return the response - formatted as a dictionary
        return {
            "bodies": [
                {
                    "id": body.id,
                    "name": body.name,
                    "master_id": body.master_id,
                    "is_surface": body.is_surface,
                }
            ]
            for body in resp.bodies
        }

    @protect_grpc
    def create_extruded_body_from_face_profile(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreateExtrudedBodyFromFaceProfileRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateExtrudedBodyFromFaceProfileRequest(
            name=kwargs["name"],
            parent=kwargs["parent_id"],
            face=kwargs["face_id"],
            distance=from_measurement_to_server_length(kwargs["distance"]) * kwargs["direction"],
        )

        # Call the gRPC service
        resp = self.stub.CreateExtrudedBodyFromFaceProfile(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def create_extruded_body_from_loft_profiles(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreateExtrudedBodyFromLoftProfilesRequest
        from ansys.api.geometry.v0.models_pb2 import TrimmedCurveList

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateExtrudedBodyFromLoftProfilesRequest(
            name=kwargs["name"],
            parent=kwargs["parent_id"],
            profiles=[
                TrimmedCurveList(
                    curves=[from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in profile]
                )
                for profile in kwargs["profiles"]
            ],
            periodic=kwargs["periodic"],
            ruled=kwargs["ruled"],
        )

        # Call the gRPC service
        resp = self.stub.CreateExtrudedBodyFromLoftProfiles(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def create_planar_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreatePlanarBodyRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreatePlanarBodyRequest(
            name=kwargs["name"],
            parent=kwargs["parent_id"],
            plane=from_plane_to_grpc_plane(kwargs["sketch"].plane),
            geometries=from_sketch_shapes_to_grpc_geometries(
                kwargs["sketch"].plane, kwargs["sketch"].edges, kwargs["sketch"].faces
            ),
        )

        # HACK: we should inform the user that NURBS curve sketches are not supported
        # prior to 26R1... and if passed, raise an error
        _nurbs_curves_compatibility(kwargs["backend_version"], request.geometries)

        # Call the gRPC service
        resp = self.stub.CreatePlanarBody(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def create_body_from_face(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreateBodyFromFaceRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateBodyFromFaceRequest(
            name=kwargs["name"],
            parent=kwargs["parent_id"],
            face=kwargs["face_id"],
        )

        # Call the gRPC service
        resp = self.stub.CreateBodyFromFace(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def create_surface_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreateSurfaceBodyRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSurfaceBodyRequest(
            name=kwargs["name"],
            parent=kwargs["parent_id"],
            trimmed_surface=from_trimmed_surface_to_grpc_trimmed_surface(kwargs["trimmed_surface"]),
        )

        # Call the gRPC service
        resp = self.stub.CreateSurfaceBody(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def create_surface_body_from_trimmed_curves(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CreateSurfaceBodyFromTrimmedCurvesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSurfaceBodyFromTrimmedCurvesRequest(
            name=kwargs["name"],
            parent=kwargs["parent_id"],
            trimmed_curves=[
                from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in kwargs["trimmed_curves"]
            ],
        )

        # Call the gRPC service
        resp = self.stub.CreateSurfaceBodyFromTrimmedCurves(request=request)

        # Return the response - formatted as a dictionary
        return {
            "id": resp.id,
            "name": resp.name,
            "master_id": resp.master_id,
            "is_surface": resp.is_surface,
        }

    @protect_grpc
    def translate(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import TranslateRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = TranslateRequest(
            ids=kwargs["ids"],
            direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
            distance=from_measurement_to_server_length(kwargs["distance"]),
        )

        # Call the gRPC service
        self.stub.Translate(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def delete(self, **kwargs) -> dict:  # noqa: D102
        # Call the gRPC service
        self.stub.Delete(request=build_grpc_id(kwargs["id"]))

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def is_suppressed(self, **kwargs) -> dict:  # noqa: D102
        # Call the gRPC service
        resp = self.stub.IsSuppressed(request=build_grpc_id(kwargs["id"]))

        # Return the response - formatted as a dictionary
        return {"result": resp.result}

    @protect_grpc
    def get_color(self, **kwargs) -> dict:  # noqa: D102
        # Call the gRPC service
        resp = self.stub.GetColor(request=build_grpc_id(kwargs["id"]))

        # Return the response - formatted as a dictionary
        return {"color": resp.color}

    @protect_grpc
    def get_faces(self, **kwargs) -> dict:  # noqa: D102
        # Call the gRPC service
        resp = self.stub.GetFaces(request=build_grpc_id(kwargs["id"]))

        # Return the response - formatted as a dictionary
        return {
            "faces": [
                {
                    "id": face.id,
                    "surface_type": face.surface_type,
                    "is_reversed": face.is_reversed,
                }
                for face in resp.faces
            ]
        }

    @protect_grpc
    def get_edges(self, **kwargs) -> dict:  # noqa: D102
        # Call the gRPC service
        resp = self.stub.GetEdges(request=build_grpc_id(kwargs["id"]))

        # Return the response - formatted as a dictionary
        return {
            "edges": [
                {
                    "id": edge.id,
                    "curve_type": edge.curve_type,
                    "is_reversed": edge.is_reversed,
                }
                for edge in resp.edges
            ]
        }

    @protect_grpc
    def get_vertices(self, **kwargs) -> dict:  # noqa: D102
        # Call the gRPC service
        resp = self.stub.GetVertices(request=build_grpc_id(kwargs["id"]))

        # Return the response - formatted as a dictionary
        return {
            "vertices": [
                {
                    "id": vertex.id.id,
                    "position": from_grpc_point_to_point3d(vertex.position),
                }
                for vertex in resp.vertices
            ]
        }

    @protect_grpc
    def get_volume(self, **kwargs) -> dict:  # noqa: D102
        # Call the gRPC service
        resp = self.stub.GetVolume(request=build_grpc_id(kwargs["id"]))

        # Return the response - formatted as a dictionary
        return {"volume": pint.Quantity(resp.volume, DEFAULT_UNITS.SERVER_VOLUME)}

    @protect_grpc
    def get_bounding_box(self, **kwargs) -> dict:  # noqa: D102
        # Call the gRPC service
        resp = self.stub.GetBoundingBox(request=build_grpc_id(kwargs["id"]))

        # Return the response - formatted as a dictionary
        return {
            "min": from_grpc_point_to_point3d(resp.box.min),
            "max": from_grpc_point_to_point3d(resp.box.max),
            "center": from_grpc_point_to_point3d(resp.box.center),
        }

    @protect_grpc
    def set_assigned_material(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import SetAssignedMaterialRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetAssignedMaterialRequest(id=kwargs["id"], material=kwargs["material"].name)

        # Call the gRPC service
        self.stub.SetAssignedMaterial(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def get_assigned_material(self, **kwargs) -> dict:  # noqa: D102
        # Call the gRPC service
        resp = self.stub.GetAssignedMaterial(request=build_grpc_id(kwargs["id"]))

        # Return the response - formatted as a dictionary
        return {"material": from_grpc_material_to_material(resp)}

    @protect_grpc
    def remove_assigned_material(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import RemoveAssignedMaterialRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = RemoveAssignedMaterialRequest(ids=[build_grpc_id(id) for id in kwargs["ids"]])

        # Call the gRPC service
        resp = self.stub.RemoveAssignedMaterial(request=request)

        # Return the response - formatted as a dictionary
        return {"successfully_removed": [id for id in resp.successfully_removed]}

    @protect_grpc
    def set_name(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import SetNameRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetNameRequest(body_id=kwargs["id"], name=kwargs["name"])

        # Call the gRPC service
        self.stub.SetName(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def set_fill_style(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import SetFillStyleRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetFillStyleRequest(body_id=kwargs["id"], fill_style=kwargs["fill_style"].value)

        # Call the gRPC service
        self.stub.SetFillStyle(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def set_suppressed(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import SetSuppressedRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetSuppressedRequest(
            bodies=[build_grpc_id(body_id) for body_id in kwargs["bodies"]],
            is_suppressed=kwargs["is_suppressed"],
        )

        # Call the gRPC service
        self.stub.SetSuppressed(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def set_color(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import SetColorRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetColorRequest(body_id=kwargs["id"], color=kwargs["color"])

        # Call the gRPC service
        self.stub.SetColor(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def rotate(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import RotateRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = RotateRequest(
            id=kwargs["id"],
            axis_origin=from_point3d_to_grpc_point(kwargs["axis_origin"]),
            axis_direction=from_unit_vector_to_grpc_direction(kwargs["axis_direction"]),
            angle=from_measurement_to_server_angle(kwargs["angle"]),
        )

        # Call the gRPC service
        self.stub.Rotate(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def scale(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import ScaleRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = ScaleRequest(
            id=kwargs["id"],
            scale=kwargs["value"],
        )

        # Call the gRPC service
        self.stub.Scale(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def mirror(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import MirrorRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = MirrorRequest(
            id=kwargs["id"],
            plane=from_plane_to_grpc_plane(kwargs["plane"]),
        )

        # Call the gRPC service
        self.stub.Mirror(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def map(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import MapRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = MapRequest(
            id=kwargs["id"],
            frame=from_frame_to_grpc_frame(kwargs["frame"]),
        )

        # Call the gRPC service
        self.stub.Map(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def get_collision(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import GetCollisionRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetCollisionRequest(
            body_1_id=kwargs["id"],
            body_2_id=kwargs["other_id"],
        )

        # Call the gRPC service
        resp = self.stub.GetCollision(request=request)

        # Return the response - formatted as a dictionary
        return {"collision_type": resp.collision}

    @protect_grpc
    def copy(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import CopyRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CopyRequest(
            id=kwargs["id"],
            parent=kwargs["parent_id"],
            name=kwargs["name"],
        )

        # Call the gRPC service
        resp = self.stub.Copy(request=request)

        # Return the response - formatted as a dictionary
        return {"master_id": resp.master_id, "name": resp.name}

    @protect_grpc
    def get_tesellation(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import GetTessellationRequest

        tess_map = {}
        resp = []  # For compatibility with stream response
        try:
            resp_single = self.stub.GetTessellation(request=build_grpc_id(kwargs["id"]))
            resp.append(resp_single)
        except grpc.RpcError as err:  # pragma: no cover
            if kwargs["backend_version"] < (25, 2, 0):
                raise err
            request = GetTessellationRequest(id=build_grpc_id(kwargs["id"]))
            resp = self.stub.GetTessellationStream(request=request)

        for elem in resp:
            for face_id, face_tess in elem.face_tessellation.items():
                tess_map[face_id] = from_grpc_tess_to_pd(face_tess)

        return {"tessellation": tess_map}

    @protect_grpc
    def get_tesellation_with_options(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import GetTessellationRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetTessellationRequest(
            id=build_grpc_id(kwargs["id"]),
            options=from_tess_options_to_grpc_tess_options(kwargs["options"]),
        )

        tess_map = {}
        resp = []  # For compatibility with stream response
        try:
            resp_single = self.stub.GetTessellationWithOptions(request)
            resp.append(resp_single)
        except grpc.RpcError:  # pragma: no cover
            resp = self.stub.GetTessellationStream(request)

        for elem in resp:
            for face_id, face_tess in elem.face_tessellation.items():
                tess_map[face_id] = from_grpc_tess_to_pd(face_tess)

        return {"tessellation": tess_map}

    @protect_grpc
    def boolean(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.bodies_pb2 import BooleanRequest

        # Call the gRPC service and build the requests accordingly
        resp = 0
        try:
            resp = self.stub.Boolean(
                request=BooleanRequest(
                    body1=kwargs["target"].id,
                    tool_bodies=[other.id for other in kwargs["other"]],
                    method=kwargs["method"],
                )
            ).empty_result
        except grpc.RpcError as err:  # pragma: no cover
            # TODO: to be deleted - old versions did not have "tool_bodies" in the request
            # This is a temporary fix to support old versions of the server - should be deleted
            # once the server is no longer supported.
            # https://github.com/ansys/pyansys-geometry/issues/1319
            if len(kwargs["other"]) > 1:
                all_resp = []
                for body2 in kwargs["other"]:
                    tmp_resp = self.stub.Boolean(
                        request=BooleanRequest(
                            body1=kwargs["target"].id,
                            body2=body2.id,
                            method=kwargs["method"],
                        )
                    ).empty_result
                    all_resp.append(tmp_resp)

                if all_resp.count(1) > 0:
                    resp = 1
            elif len(kwargs["other"]) == 1:
                resp = self.stub.Boolean(
                    request=BooleanRequest(
                        body1=kwargs["target"].id,
                        body2=kwargs["other"][0].id,
                        method=kwargs["method"],
                    )
                ).empty_result
            else:
                raise err

        if resp == 1:
            raise ValueError(
                f"Boolean operation of type '{kwargs['method']}' failed: {kwargs['err_msg']}.\n"
                f"Involving bodies:{kwargs['target']}, {kwargs['other']}"
            )

        # Return the response - formatted as a dictionary
        return {}



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for interacting with the Geometry service."""

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Generator, Optional

from grpc import Channel

from ansys.api.geometry.v0.commands_pb2 import UploadFileRequest
from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
from ansys.geometry.core.connection.backend import ApiVersions, BackendType
from ansys.geometry.core.connection.client import GrpcClient
import ansys.geometry.core.connection.defaults as pygeom_defaults
from ansys.geometry.core.errors import GeometryRuntimeError, protect_grpc
from ansys.geometry.core.misc.checks import check_type, deprecated_method, min_backend_version
from ansys.geometry.core.misc.options import ImportOptions
from ansys.geometry.core.tools.measurement_tools import MeasurementTools
from ansys.geometry.core.tools.prepare_tools import PrepareTools
from ansys.geometry.core.tools.repair_tools import RepairTools
from ansys.geometry.core.tools.unsupported import UnsupportedCommands
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.connection.docker_instance import LocalDockerInstance
    from ansys.geometry.core.connection.product_instance import ProductInstance
    from ansys.geometry.core.designer.design import Design
    from ansys.geometry.core.designer.geometry_commands import GeometryCommands
    from ansys.platform.instancemanagement import Instance


class Modeler:
    """Provides for interacting with an open session of the Geometry service.

    Parameters
    ----------
    host : str,  default: DEFAULT_HOST
        Host where the server is running.
    port : str | int, default: DEFAULT_PORT
        Port number where the server is running.
    channel : ~grpc.Channel, default: None
        gRPC channel for server communication.
    remote_instance : ansys.platform.instancemanagement.Instance, default: None
        Corresponding remote instance when the Geometry service
        is launched using `PyPIM <https://github.com/ansys/pypim>`_. This instance
        is deleted when the :func:`GrpcClient.close <ansys.geometry.core.client.GrpcClient.close>`
        method is called.
    docker_instance : LocalDockerInstance, default: None
        Corresponding local Docker instance when the Geometry service is launched using the
        :func:`launch_docker_modeler<ansys.geometry.core.connection.launcher.launch_docker_modeler>`
        method. This instance is deleted when the
        :func:`GrpcClient.close <ansys.geometry.core.client.GrpcClient.close>`
        method is called.
    product_instance : ProductInstance, default: None
        Corresponding local product instance when the product (Discovery or SpaceClaim)
        is launched through the ``launch_modeler_with_geometry_service()``,
        ``launch_modeler_with_discovery()`` or the ``launch_modeler_with_spaceclaim()``
        interface. This instance will be deleted
        when the :func:`GrpcClient.close <ansys.geometry.core.client.GrpcClient.close >`
        method is called.
    timeout : Real, default: 120
        Time in seconds for trying to achieve the connection.
    logging_level : int, default: INFO
        Logging level to apply to the client.
    logging_file : str, Path, default: None
        File to output the log to, if requested.
    """

    def __init__(
        self,
        host: str = pygeom_defaults.DEFAULT_HOST,
        port: str | int = pygeom_defaults.DEFAULT_PORT,
        channel: Channel | None = None,
        remote_instance: Optional["Instance"] = None,
        docker_instance: Optional["LocalDockerInstance"] = None,
        product_instance: Optional["ProductInstance"] = None,
        timeout: Real = 120,
        logging_level: int = logging.INFO,
        logging_file: Path | str | None = None,
    ):
        """Initialize the ``Modeler`` class."""
        from ansys.geometry.core.designer.geometry_commands import GeometryCommands

        self._grpc_client = GrpcClient(
            host=host,
            port=port,
            channel=channel,
            remote_instance=remote_instance,
            docker_instance=docker_instance,
            product_instance=product_instance,
            timeout=timeout,
            logging_level=logging_level,
            logging_file=logging_file,
        )

        # Single design for the Modeler
        self._design: Optional["Design"] = None

        # Enabling tools/commands for all: repair and prepare tools, geometry commands
        self._measurement_tools = MeasurementTools(self._grpc_client, _internal_use=True)
        self._repair_tools = RepairTools(self._grpc_client, self, _internal_use=True)
        self._prepare_tools = PrepareTools(self._grpc_client, _internal_use=True)
        self._geometry_commands = GeometryCommands(self._grpc_client, _internal_use=True)
        self._unsupported = UnsupportedCommands(self._grpc_client, self, _internal_use=True)

    @property
    def client(self) -> GrpcClient:
        """``Modeler`` instance client."""
        return self._grpc_client

    @property
    def design(self) -> "Design":
        """Retrieve the design within the modeler workspace."""
        return self._design

    @property
    @deprecated_method(alternative="design", version="0.9.0", remove="0.12.0")
    def designs(self) -> dict[str, "Design"]:
        """Retrieve the design within the modeler workspace.

        Notes
        -----
        This method is deprecated. Use the :func:`design` property instead.
        """
        return {self._design.id: self._design}

    def create_design(self, name: str) -> "Design":
        """Initialize a new design with the connected client.

        Parameters
        ----------
        name : str
            Name for the new design.

        Returns
        -------
        Design
            Design object created on the server.
        """
        from ansys.geometry.core.designer.design import Design

        check_type(name, str)

        # If a previous design was available... inform the user that it will be deleted
        # when creating a new design.
        if self._design is not None and self._design.is_active:
            self.client.log.warning("Closing previous design before creating a new one.")
            self._design.close()

        # Create the new design
        design = Design(name, self)

        # Update the design stored in the modeler
        self._design = design

        return self._design

    def get_active_design(self, sync_with_backend: bool = True) -> "Design":
        """Get the active design on the modeler object.

        Parameters
        ----------
        sync_with_backend : bool, default: True
            Whether to sync the active design with the remote service. If set to False,
            the active design may be out-of-sync with the remote service. This is useful
            when the active design is known to be up-to-date.

        Returns
        -------
        Design
            Design object already existing on the modeler.
        """
        if self._design is not None and self._design.is_active:
            # Check if sync_with_backend is requested
            if sync_with_backend:
                self._design._update_design_inplace()
            return self._design
        else:
            self.client.log.warning("No active design available.")

        return None

    def read_existing_design(self) -> "Design":
        """Read the existing design on the service with the connected client.

        Returns
        -------
        Design
            Design object already existing on the server.
        """
        from ansys.geometry.core.designer.design import Design

        # Simply deactivate the existing design in case it is active...
        # - If it is the same design, it will be re-read (but we should not close it on the server)
        # - If it is a different design, it was closed previously (via open_file or create_design)
        if self._design is not None and self._design.is_active:
            self._design._is_active = False

        self._design = Design("", self, read_existing_design=True)

        return self._design

    def close(self, close_design: bool = True) -> None:
        """Access the client's close method.

        Parameters
        ----------
        close_design : bool, default: True
            Whether to close the design before closing the client.
        """
        # Close design (if requested)
        if close_design and self._design is not None:
            self._design.close()

        # Close the client
        self.client.close()

    def exit(self, close_design: bool = True) -> None:
        """Access the client's close method.

        Parameters
        ----------
        close_design : bool, default: True
            Whether to close the design before closing the client.

        Notes
        -----
        This method is calling the same method as
        :func:`close() <ansys.geometry.core.modeler.Modeler.close>`.
        """
        self.close(close_design=close_design)

    @protect_grpc
    def _upload_file(
        self,
        file_path: str,
        open_file: bool = False,
        import_options: ImportOptions = ImportOptions(),
    ) -> str:
        """Upload a file from the client to the server.

        Parameters
        ----------
        file_path : str
            Path of the file to upload. The extension of the file must be included.
        open_file : bool, default: False
            Whether to open the file in the Geometry service.
        import_options : ImportOptions
            Import options that toggle certain features when opening a file.

        Returns
        -------
        file_path : str
            Full path of the file uploaded to the server.

        Notes
        -----
        This method creates a file on the server that has the same name and extension
        as the file on the client.
        """
        from pathlib import Path

        fp_path = Path(file_path).resolve()

        if not fp_path.exists():
            raise ValueError(f"Could not find file: {file_path}")
        if fp_path.is_dir():
            raise ValueError("File path must lead to a file, not a directory.")

        file_name = fp_path.name

        with fp_path.open(mode="rb") as file:
            data = file.read()

        c_stub = CommandsStub(self.client.channel)

        response = c_stub.UploadFile(
            UploadFileRequest(
                data=data,
                file_name=file_name,
                open=open_file,
                import_options=import_options.to_dict(),
            )
        )
        return response.file_path

    @protect_grpc
    def _upload_file_stream(
        self,
        file_path: str,
        open_file: bool = False,
        import_options: ImportOptions = ImportOptions(),
    ) -> str:
        """Upload a file from the client to the server via streaming.

        Parameters
        ----------
        file_path : str
            Path of the file to upload. The extension of the file must be included.
        open_file : bool, default: False
            Whether to open the file in the Geometry service.
        import_options : ImportOptions
            Import options that toggle certain features when opening a file.

        Returns
        -------
        file_path : str
            Full path of the file uploaded to the server.

        Notes
        -----
        This method creates a file on the server that has the same name and extension
        as the file on the client.
        """
        from pathlib import Path

        fp_path = Path(file_path).resolve()

        if not fp_path.exists():
            raise ValueError(f"Could not find file: {file_path}")
        if fp_path.is_dir():
            raise ValueError("File path must lead to a file, not a directory.")

        c_stub = CommandsStub(self.client.channel)

        response = c_stub.StreamFileUpload(
            self._generate_file_chunks(fp_path, open_file, import_options)
        )
        return response.file_path

    def _generate_file_chunks(
        self, file_path: Path, open_file: bool, import_options: ImportOptions
    ) -> Generator[UploadFileRequest, None, None]:
        """Generate appropriate chunk sizes for uploading files.

        Parameters
        ----------
        file_path : Path
            Path of the file to upload. The extension of the file must be included.
        open_file : bool
            Whether to open the file in the Geometry service.
        import_options : ImportOptions
            Import options that toggle certain features when opening a file.

        Yields
        ------
        UploadFileRequest
            Request object for uploading a file in chunks.
        """
        msg_buffer = 5 * 1024  # 5KB - for additional message data
        if pygeom_defaults.MAX_MESSAGE_LENGTH - msg_buffer < 0:  # pragma: no cover
            raise ValueError("MAX_MESSAGE_LENGTH is too small for file upload.")

        chunk_size = pygeom_defaults.MAX_MESSAGE_LENGTH - msg_buffer
        with Path.open(file_path, "rb") as file:
            while chunk := file.read(chunk_size):
                yield UploadFileRequest(
                    data=chunk,
                    file_name=file_path.name,
                    open=open_file,
                    import_options=import_options.to_dict(),
                )

    def open_file(
        self,
        file_path: str | Path,
        upload_to_server: bool = True,
        import_options: ImportOptions = ImportOptions(),
    ) -> "Design":
        """Open a file.

        This method imports a design into the service. On Windows and Linux, ``.scdocx``, ``.dsco``,
        and reader formats are supported. Please see notes for supported reader formats.

        If the file is a shattered assembly with external references, the whole containing folder
        will need to be uploaded. Ensure proper folder structure in order to prevent the uploading
        of unnecessary files.

        Parameters
        ----------
        file_path : str, ~pathlib.Path
            Path of the file to open. The extension of the file must be included.
        upload_to_server : bool
            True if the service is running on a remote machine. If service is running on the local
            machine, set to False, as there is no reason to upload the file.
        import_options : ImportOptions
            Import options that toggle certain features when opening a file.

        Returns
        -------
        Design
            Newly imported design.

        Notes
        -----
        Format and latest supported version
            * AutoCAD 2024
            * CATIA V5 2024
            * CATIA V6 2024
            * Creo Parametric 11
            * IGES 5.3
            * Inventor 2025
            * JT 10.10
            * NX 2412
            * Rhino 8
            * Solid Edge 2025
            * SOLIDWORKS 2025
            * STEP AP242
        """
        # Use str format of Path object here
        file_path = str(file_path) if isinstance(file_path, Path) else file_path

        # Close the existing design if it is active
        if self._design is not None and self._design.is_active:
            self._design.close()

        # Format-specific logic - upload the whole containing folder for assemblies
        if upload_to_server:
            fp_path = Path(file_path)
            file_size_kb = fp_path.stat().st_size
            if any(
                ext in str(file_path) for ext in [".CATProduct", ".asm", ".solution", ".sldasm"]
            ):
                dir = fp_path.parent
                for file in dir.iterdir():
                    full_path = file.resolve()
                    if full_path != fp_path:
                        if full_path.stat().st_size < pygeom_defaults.MAX_MESSAGE_LENGTH:
                            self._upload_file(full_path)
                        elif self.client.backend_version >= (25, 2, 0):
                            self._upload_file_stream(full_path)
                        else:  # pragma: no cover
                            raise RuntimeError(
                                "File is too large to upload."
                                " Service versions above 25R2 support streaming."
                            )

            if file_size_kb < pygeom_defaults.MAX_MESSAGE_LENGTH:
                self._upload_file(file_path, True, import_options)
            elif self.client.backend_version >= (25, 2, 0):
                self._upload_file_stream(file_path, True, import_options)
            else:  # pragma: no cover
                raise RuntimeError(
                    "File is too large to upload. Service versions above 25R2 support streaming."
                )
        else:
            self.client.services.designs.open(
                filepath=file_path,
                import_options=import_options,
            )

        return self.read_existing_design()

    def __repr__(self) -> str:
        """Represent the modeler as a string."""
        lines = []
        lines.append(f"Ansys Geometry Modeler ({hex(id(self))})")
        lines.append("")
        lines.append(str(self.client))
        return "\n".join(lines)

    @protect_grpc
    def run_discovery_script_file(
        self,
        file_path: str | Path,
        script_args: dict[str, str] | None = None,
        import_design: bool = False,
        api_version: int | str | ApiVersions = None,
    ) -> tuple[dict[str, str], Optional["Design"]]:
        """Run a Discovery script file.

        .. note::

            If arguments are passed to the script, they must be in the form of a dictionary.
            On the server side, the script will receive the arguments as a dictionary of strings,
            under the variable name ``argsDict``. For example, if the script is called with the
            arguments ``run_discovery_script_file(..., script_args = {"length": "20"}, ...)``,
            the script will receive the dictionary ``argsDict`` with the key-value pair
            ``{"length": "20"}``.

        .. note::

            If an output is expected from the script, it will be returned as a dictionary of
            strings. The keys and values of the dictionary are the variables and their values
            that the script returns. However, it is necessary that the script creates a
            dictionary called ``result`` with the variables and their values that are expected
            to be returned. For example, if the script is expected to return the number of bodies
            in the design, the script should create a dictionary called ``result`` with the
            key-value pair ``{"numBodies": numBodies}``, where ``numBodies`` is the number of
            bodies in the design.

        The implied API version of the script should match the API version of the running
        Geometry Service. DMS API versions 24.1 and later are supported. DMS is a
        Windows-based modeling service that has been containerized to ease distribution,
        execution, and remotability operations.

        Parameters
        ----------
        file_path : str | ~pathlib.Path
            Path of the file. The extension of the file must be included.
        script_args : dict[str, str], optional.
            Arguments to pass to the script. By default, ``None``.
        import_design : bool, optional.
            Whether to refresh the current design from the service. When the script
            is expected to modify the existing design, set this to ``True`` to retrieve
            up-to-date design data. When this is set to ``False`` (default) and the
            script modifies the current design, the design may be out-of-sync. By default,
            ``False``.
        api_version : int | str | ApiVersions, optional
            The scripting API version to use. For example, version 24.1 can be passed as
            an integer 241, a string "241" or using the
            ``ansys.geometry.core.connection.backend.ApiVersions`` enum class.
            By default, ``None``. When specified, the service will attempt to run the script with
            the specified API version. If the API version is not supported, the service will raise
            an error. If you are using Discovery or SpaceClaim, the product will determine the API
            version to use, so there is no need to specify this parameter.

        Returns
        -------
        dict[str, str]
            Values returned from the script.
        Design, optional
            Up-to-date current design. This is only returned if ``import_design=True``.

        Raises
        ------
        GeometryRuntimeError
            If the Discovery script fails to run. Otherwise, assume that the script
            ran successfully.

        Notes
        -----
            The Ansys Geometry Service only supports scripts that are of the
            same version as the running service. Any ``api_version`` input will
            be ignored.
        """
        # Use str format of Path object here
        file_path = str(file_path) if isinstance(file_path, Path) else file_path

        # Check if API version is specified... if so, validate it
        if api_version is not None:
            if BackendType.is_headless_service(self.client.backend_type):
                self.client.log.warning(
                    "The Ansys Geometry Service only supports "
                    "scripts that are of its same API version."
                )
                self.client.log.warning("Ignoring specified API version.")
                api_version = None
            else:  # pragma: no cover
                # Testing is only performed on Windows Service...
                # but this method has been tested independently
                api_version = ApiVersions.parse_input(api_version)

        serv_path = self._upload_file(file_path)

        self.client.log.debug(f"Running Discovery script file at {file_path}...")
        response = self.client.services.dbu_application.run_script(
            script_path=serv_path,
            script_args=script_args,
            api_version=api_version.value if api_version is not None else None,
        )

        if not response.get("success"):
            raise GeometryRuntimeError(response.get("message"))

        self.client.log.debug(f"Script result message: {response.get('message')}")

        if import_design:
            return response.get("values"), self.read_existing_design()
        else:
            return response.get("values"), None

    @property
    def repair_tools(self) -> RepairTools:
        """Access to repair tools."""
        return self._repair_tools

    @property
    def prepare_tools(self) -> PrepareTools:
        """Access to prepare tools."""
        return self._prepare_tools

    @property
    @min_backend_version(24, 2, 0)
    def measurement_tools(self) -> MeasurementTools:
        """Access to measurement tools.

        Notes
        -----
        This property is only available starting on Ansys release 24R2.
        """
        return self._measurement_tools

    @property
    def geometry_commands(self) -> "GeometryCommands":
        """Access to geometry commands."""
        return self._geometry_commands

    @property
    def unsupported(self) -> "UnsupportedCommands":
        """Access to unsupported commands."""
        return self._unsupported

    @min_backend_version(25, 1, 0)
    def get_service_logs(
        self,
        all_logs: bool = False,
        dump_to_file: bool = False,
        logs_folder: str | Path | None = None,
    ) -> str | dict[str, str] | Path:
        """Get the service logs.

        Parameters
        ----------
        all_logs : bool, default: False
            Flag indicating whether all logs should be retrieved. By default,
            only the current logs are retrieved.
        dump_to_file : bool, default: False
            Flag indicating whether the logs should be dumped to a file.
            By default, the logs are not dumped to a file.
        logs_folder : str,  Path or None, default: None
            Name of the folder where the logs should be dumped. This parameter
            is only used if the ``dump_to_file`` parameter is set to ``True``.

        Returns
        -------
        str
            Service logs as a string. This is returned if the ``dump_to_file`` parameter
            is set to ``False``.
        dict[str, str]
            Dictionary containing the logs. The keys are the logs names,
            and the values are the logs as strings. This is returned if the ``all_logs``
            parameter is set to ``True`` and the ``dump_to_file`` parameter
            is set to ``False``.
        Path
            Path to the folder containing the logs (if the ``all_logs``
            parameter is set to ``True``) or the path to the log file (if only
            the current logs are retrieved). The ``dump_to_file`` parameter
            must be set to ``True``.

        Notes
        -----
        This property is only available starting on Ansys release 25R1.
        """
        return self.client._get_service_logs(
            all_logs=all_logs, dump_to_file=dump_to_file, logs_folder=logs_folder
        )



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides tools for measurement."""

from typing import TYPE_CHECKING, Union

from ansys.geometry.core.connection import GrpcClient
from ansys.geometry.core.errors import GeometryRuntimeError
from ansys.geometry.core.misc.checks import min_backend_version
from ansys.geometry.core.misc.measurements import Distance

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face


class Gap:
    """Represents a gap between two bodies.

    Parameters
    ----------
    distance : Distance
        Distance between two sides of the gap.
    """

    def __init__(self, distance: Distance):
        """Initialize ``Gap`` class."""
        self._distance = distance

    @property
    def distance(self) -> Distance:
        """Returns the closest distance between two bodies."""
        return self._distance


class MeasurementTools:
    """Measurement tools for PyAnsys Geometry.

    Parameters
    ----------
    grpc_client : GrpcClient
        gRPC client to use for the measurement tools.
    _internal_use : bool, optional
        Internal flag to prevent direct instantiation by users.
        This parameter is for internal use only.

    Raises
    ------
    GeometryRuntimeError
        If the class is instantiated directly by users instead
        of through the modeler.

    Notes
    -----
    This class should not be instantiated directly. Use
    ``modeler.measurement_tools`` instead.
    """

    def __init__(self, grpc_client: GrpcClient, _internal_use: bool = False):
        """Initialize measurement tools class."""
        if not _internal_use:
            raise GeometryRuntimeError(
                "MeasurementTools should not be instantiated directly. "
                "Use 'modeler.measurement_tools' to access measurement tools."
            )
        self._grpc_client = grpc_client

    @min_backend_version(24, 2, 0)
    def min_distance_between_objects(
        self, object1: Union["Body", "Face", "Edge"], object2: Union["Body", "Face", "Edge"]
    ) -> Gap:
        """Find the gap between two bodies.

        Parameters
        ----------
        object1 : Union[Body, Face, Edge]
            First object to measure the gap.
        object2 : Union[Body, Face, Edge]
            Second object to measure the gap.

        Returns
        -------
        Gap
            Gap between two bodies.

        Warnings
        --------
        This method is only available starting on Ansys release 24R2.
        Also, using Face and Edge objects as inputs requires a minimum Ansys
        release of 25R2.
        """
        from ansys.geometry.core.designer.edge import Edge
        from ansys.geometry.core.designer.face import Face

        # Face and Edge objects are only accepted if the backend version is 25R2 or later.
        for obj in (object1, object2):
            if isinstance(obj, (Face, Edge)) and self._grpc_client.backend_version < (25, 2, 0):
                raise GeometryRuntimeError(
                    f"The method '{self.min_distance_between_objects.__name__}' using "
                    "Face and Edge objects as inputs requires a minimum Ansys release version of "
                    f"25.2.0, but the current version used is {self._grpc_client.backend_version}."
                )

        response = self._grpc_client.services.measurement_tools.min_distance_between_objects(
            selection=[object1.id, object2.id],
            backend_version=self._grpc_client.backend_version,
        )
        return Gap(response.get("distance"))



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for managing an edge."""

from enum import Enum, unique
from typing import TYPE_CHECKING

from pint import Quantity

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.errors import GeometryRuntimeError
from ansys.geometry.core.math.bbox import BoundingBox
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.misc.checks import ensure_design_is_active, min_backend_version
from ansys.geometry.core.shapes.curves.trimmed_curve import ReversedTrimmedCurve, TrimmedCurve
from ansys.geometry.core.shapes.parameterization import Interval

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.face import Face
    from ansys.geometry.core.designer.vertex import Vertex


@unique
class CurveType(Enum):
    """Provides values for the curve types supported."""

    CURVETYPE_UNKNOWN = 0
    CURVETYPE_LINE = 1
    CURVETYPE_CIRCLE = 2
    CURVETYPE_ELLIPSE = 3
    CURVETYPE_NURBS = 4
    CURVETYPE_PROCEDURAL = 5


class Edge:
    """Represents a single edge of a body within the design assembly.

    This class synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    curve_type : CurveType
        Type of curve that the edge forms.
    body : Body
        Parent body that the edge constructs.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    is_reversed : bool
        Direction of the edge.
    """

    def __init__(
        self,
        id: str,
        curve_type: CurveType,
        body: "Body",
        grpc_client: GrpcClient,
        is_reversed: bool = False,
    ):
        """Initialize the ``Edge`` class."""
        self._id = id
        self._curve_type = curve_type
        self._body = body
        self._grpc_client = grpc_client
        self._is_reversed = is_reversed
        self._shape = None

    @property
    def id(self) -> str:
        """ID of the edge."""
        return self._id

    @property
    def _grpc_id(self) -> EntityIdentifier:
        """Entity ID of this edge on the server side."""
        return EntityIdentifier(id=self._id)

    @property
    def body(self) -> "Body":
        """Body of the edge."""
        return self._body

    @property
    def is_reversed(self) -> bool:
        """Flag indicating if the edge is reversed."""
        return self._is_reversed

    @property
    @ensure_design_is_active
    @min_backend_version(24, 2, 0)
    def shape(self) -> TrimmedCurve:
        """Underlying trimmed curve of the edge.

        If the edge is reversed, its shape is the ``ReversedTrimmedCurve`` type, which swaps the
        start and end points of the curve and handles parameters to allow evaluation as if the
        curve is not reversed.

        Warnings
        --------
        This method is only available starting on Ansys release 24R2.
        """
        if self._shape is None:
            self._grpc_client.log.debug("Requesting edge properties from server.")
            geometry = self._grpc_client.services.edges.get_curve(id=self._id).get("curve")

            response = self._grpc_client.services.edges.get_start_and_end_points(id=self._id)
            start = response.get("start")
            end = response.get("end")

            length = self._grpc_client.services.edges.get_length(id=self._id).get("length")

            response = self._grpc_client.services.edges.get_interval(id=self._id)
            interval = Interval(response.get("start"), response.get("end"))

            self._shape = (
                ReversedTrimmedCurve(geometry, start, end, interval, length.value)
                if self.is_reversed
                else TrimmedCurve(geometry, start, end, interval, length.value)
            )
        return self._shape

    @property
    @ensure_design_is_active
    def length(self) -> Quantity:
        """Calculated length of the edge."""
        try:
            return self.shape.length
        except GeometryRuntimeError:  # pragma: no cover
            # Only for versions earlier than 24.2.0 (before the introduction of the shape property)
            self._grpc_client.log.debug("Requesting edge length from server.")
            return self._grpc_client.services.edges.get_length(id=self._id).get("length").value

    @property
    def curve_type(self) -> CurveType:
        """Curve type of the edge."""
        return self._curve_type

    @property
    @ensure_design_is_active
    def faces(self) -> list["Face"]:
        """Faces that contain the edge."""
        from ansys.geometry.core.designer.face import Face, SurfaceType

        self._grpc_client.log.debug("Requesting edge faces from server.")
        response = self._grpc_client.services.edges.get_faces(id=self._id)
        return [
            Face(
                face_resp.get("id"),
                SurfaceType(face_resp.get("surface_type")),
                self._body,
                self._grpc_client,
                face_resp.get("is_reversed"),
            )
            for face_resp in response.get("faces")
        ]

    @property
    @ensure_design_is_active
    @min_backend_version(26, 1, 0)
    def vertices(self) -> list["Vertex"]:
        """Vertices that define the edge."""
        from ansys.geometry.core.designer.vertex import Vertex

        self._grpc_client.log.debug("Requesting edge vertices from server.")
        response = self._grpc_client.services.edges.get_vertices(id=self._id)

        return [
            Vertex(
                vertex_resp.get("id"),
                vertex_resp.get("position"),
            )
            for vertex_resp in response.get("vertices")
        ]

    @property
    @ensure_design_is_active
    def start(self) -> Point3D:
        """Start point of the edge."""
        try:
            return self.shape.start
        except GeometryRuntimeError:  # pragma: no cover
            # Only for versions earlier than 24.2.0 (before the introduction of the shape property)
            self._grpc_client.log.debug("Requesting edge start point from server.")
            return self._grpc_client.services.edges.get_start_and_end_points(id=self._id).get(
                "start"
            )

    @property
    @ensure_design_is_active
    def end(self) -> Point3D:
        """End point of the edge."""
        try:
            return self.shape.end
        except GeometryRuntimeError:  # pragma: no cover
            # Only for versions earlier than 24.2.0 (before the introduction of the shape property)
            self._grpc_client.log.debug("Requesting edge end point from server.")
            return self._grpc_client.services.edges.get_start_and_end_points(id=self._id).get("end")

    @property
    @ensure_design_is_active
    @min_backend_version(25, 2, 0)
    def bounding_box(self) -> BoundingBox:
        """Bounding box of the edge.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        self._grpc_client.log.debug("Requesting bounding box from server.")

        response = self._grpc_client.services.edges.get_bounding_box(id=self._id)
        return BoundingBox(
            response.get("min_corner"), response.get("max_corner"), response.get("center")
        )



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Auxiliary functions for the PyAnsys Geometry library."""

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.beam import Beam
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.component import Component
    from ansys.geometry.core.designer.design import Design
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face
    from ansys.geometry.core.designer.vertex import Vertex

try:
    from ansys.tools.visualization_interface.utils.color import Color

    DEFAULT_COLOR = Color.DEFAULT.value
except ModuleNotFoundError:
    DEFAULT_COLOR = "#D6F7D1"


def get_design_from_component(component: "Component") -> "Design":
    """Get the ``Design`` of the given ``Component`` object.

    Parameters
    ----------
    component : Component
        The component object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided component object.
    """
    # Get the design of the component
    while component.parent_component is not None:
        component = component.parent_component

    # Return the design
    return component


def get_design_from_body(body: "Body") -> "Design":
    """Get the ``Design`` of the given ``Body`` object.

    Parameters
    ----------
    body : Body
        The body object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided body object.
    """
    # Get the parent component of the body
    component = body.parent_component

    # Get the design of the component
    return get_design_from_component(component)


def get_design_from_face(face: "Face") -> "Design":
    """Get the ``Design`` of the given ``Face`` object.

    Parameters
    ----------
    face : Face
        The face object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided face object.
    """
    # Get the parent body of the face
    body = face.body

    # Get the design of the body
    return get_design_from_body(body)


def get_design_from_edge(edge: "Edge") -> "Design":
    """Get the ``Design`` of the given ``Edge`` object.

    Parameters
    ----------
    edge : Edge
        The edge object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided edge object.
    """
    # Get the one of the bodies of the edge
    body = edge.faces[0].body

    # Get the design of the body
    return get_design_from_body(body)


def __traverse_component_elem(elem: str, comp: Union["Design", "Component"]) -> list:
    """Traverses all elements of a component or design, given its name."""
    elems = []
    comp_elems = getattr(comp, elem)
    elems.extend(comp_elems)
    for component in comp.components:
        elems.extend(__traverse_component_elem(elem, component))

    return elems


def __traverse_all_bodies(comp: Union["Design", "Component"]) -> list["Body"]:
    """Traverse all bodies in a design/component and all its subcomponents."""
    return __traverse_component_elem("bodies", comp)


def __traverse_all_beams(comp: Union["Design", "Component"]) -> list["Body"]:
    """Traverse all beams in a design/component and all its subcomponents."""
    return __traverse_component_elem("beams", comp)


def get_all_bodies_from_design(design: "Design") -> list["Body"]:
    """Find all the ``Body`` objects inside a ``Design``.

    Parameters
    ----------
    design : Design
        Parent design for the bodies.

    Returns
    -------
    list[Body]
        List of Body objects.

    Notes
    -----
    This method takes a design and gets the corresponding ``Body`` objects.
    """
    return __traverse_all_bodies(design)


def get_bodies_from_ids(design: "Design", body_ids: list[str]) -> list["Body"]:
    """Find the ``Body`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the faces.
    body_ids : list[str]
        List of body ids.

    Returns
    -------
    list[Body]
        List of Body objects.

    Notes
    -----
    This method takes a design and body ids, and gets their corresponding ``Body`` object.
    """
    return [body for body in __traverse_all_bodies(design) if body.id in body_ids]


def get_components_from_ids(design: "Design", component_ids: list[str]) -> list["Component"]:
    """Find the ``Component`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the components.
    component_ids : list[str]
        List of component ids.

    Returns
    -------
    list[Component]
        List of Component objects.

    Notes
    -----
    This method takes a design and component ids, and gets their corresponding ``Component`` object.
    """
    return [
        comp for comp in __traverse_component_elem("components", design) if comp.id in component_ids
    ]  # noqa: E501


def get_faces_from_ids(design: "Design", face_ids: list[str]) -> list["Face"]:
    """Find the ``Face`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the faces.
    face_ids : list[str]
        List of face ids.

    Returns
    -------
    list[Face]
        List of Face objects.

    Notes
    -----
    This method takes a design and face ids, and gets their corresponding ``Face`` object.
    """
    return [
        face for body in __traverse_all_bodies(design) for face in body.faces if face.id in face_ids
    ]  # noqa: E501


def get_edges_from_ids(design: "Design", edge_ids: list[str]) -> list["Edge"]:
    """Find the ``Edge`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the edges.
    edge_ids : list[str]
        List of edge ids.

    Returns
    -------
    list[Edge]
        List of Edge objects.

    Notes
    -----
    This method takes a design and edge ids, and gets their corresponding ``Edge`` objects.
    """
    return [
        edge for body in __traverse_all_bodies(design) for edge in body.edges if edge.id in edge_ids
    ]  # noqa: E501


def get_vertices_from_ids(design: "Design", vertex_ids: list[str]) -> list["Vertex"]:
    """Find the ``Vertex`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the vertices.
    vertex_ids : list[str]
        List of vertex ids.

    Returns
    -------
    list[Vertex]
        List of Vertex objects.

    Notes
    -----
    This method takes a design and vertex ids, and gets their corresponding ``Vertex`` objects.
    """
    return [
        vertex
        for body in __traverse_all_bodies(design)
        for vertex in body.vertices
        if vertex.id in vertex_ids
    ]  # noqa: E501


def get_beams_from_ids(design: "Design", beam_ids: list[str]) -> list["Beam"]:
    """Find the ``Beam`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the beams.
    beam_ids : list[str]
        List of beam ids.

    Returns
    -------
    list[Beam]
        List of Beam objects.

    Notes
    -----
    This method takes a design and beam ids, and gets their corresponding ``Beam`` objects.
    """
    return [beam for beam in __traverse_all_beams(design) if beam.id in beam_ids]  # noqa: E501


def convert_color_to_hex(
    color: str | tuple[float, float, float] | tuple[float, float, float, float],
) -> str:
    """Get the hex string color from input formats.

    Parameters
    ----------
    color : str | tuple[float, float, float] | tuple[float, float, float, float]
        Color to set the body to. This can be a string representing a color name
        or a tuple of RGB values in the range [0, 1] (RGBA) or [0, 255] (pure RGB).

    Returns
    -------
    str
        The hex code string for the color, formatted #rrggbbaa.
    """
    import matplotlib.colors as mcolors

    try:
        if isinstance(color, tuple):
            # Ensure that all elements are within 0-1 or 0-255 range
            if all(0 <= c <= 1 for c in color):
                # Ensure they are floats if in 0-1 range
                if not all(isinstance(c, float) for c in color):
                    raise ValueError("RGB values in the 0-1 range must be floats.")
            elif all(0 <= c <= 255 for c in color):
                # Ensure they are integers if in 0-255 range
                if not all(isinstance(c, int) for c in color):
                    raise ValueError("RGB values in the 0-255 range must be integers.")
                # Normalize the 0-255 range to 0-1
                color = tuple(c / 255.0 for c in color)
            else:
                raise ValueError("RGB tuple contains mixed ranges or invalid values.")

            color = mcolors.to_hex(color, keep_alpha=True)
        elif isinstance(color, str):
            color = mcolors.to_hex(color, keep_alpha=True)
    except ValueError as err:
        raise ValueError(f"Invalid color value: {err}")

    return color


def convert_opacity_to_hex(opacity: float) -> str:
    """Get the hex string from an opacity value.

    Parameters
    ----------
    opacity : float
        Opacity to set body to. Must be in the range [0, 1].

    Returns
    -------
    The hex code for the opacity formatted #aa
    """
    try:
        # Ensure that the value is within 0-1 range
        if 0 <= opacity <= 1:
            return "{:02x}".format(int(opacity * 255))
        else:
            raise ValueError("Opacity value must be between 0 and 1.")
    except ValueError as err:
        raise ValueError(f"Invalid color value: {err}")
